import argparse
import yucca
from yucca.utils.task_ids import maybe_get_task_from_task_id
from yucca.utils.files_and_folders import recursive_find_python_class
from batchgenerators.utilities.file_and_folder_operations import join


def main():
    parser = argparse.ArgumentParser()

    # Required Arguments #
    parser.add_argument(
        "-t",
        "--task",
        help="Name of the task used for training. "
        "The data should already be preprocessed using yucca_preprocess"
        "Argument should be of format: TaskXXX_MYTASK",
    )

    # Optional arguments with default values #
    parser.add_argument(
        "-m",
        help="Model Architecture. Should be one of MultiResUNet or UNet"
        " Note that this is case sensitive. "
        "Defaults to the standard UNet.",
        default="UNet",
    )
    parser.add_argument(
        "-d",
        help="Dimensionality of the Model. Can be 3D or 2D. "
        "Defaults to 3D. Note that this will always be 2D if ensemble is enabled.",
    )
    parser.add_argument("-tr", help="Trainer Class to be used. " "Defaults to the basic YuccaTrainer", default="YuccaTrainer")
    parser.add_argument(
        "-pl",
        help="Plan ID to be used. "
        "This specifies which plan and preprocessed data to use for training "
        "on the given task. Defaults to the YuccaPlanne folder",
        default="YuccaPlanner",
    )
    parser.add_argument(
        "-f",
        help="Fold to use for training. Unless manually assigned, "
        "folds [0,1,2,3,4] will be created automatically. "
        "Defaults to training on fold 0",
        default=0,
    )
    parser.add_argument(
        "--ensemble",
        help="Used to train ensemble/2.5D models. Will run 3 consecutive trainings.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--fast", help="Used to speed up training, possibly at the expense of performance", default=False, action="store_true"
    )

    # The following can be changed to run training with alternative LR, Loss and/or Momentum ###
    parser.add_argument(
        "--lr",
        help="Should only be used to employ alternative Learning Rate. " "Format should be scientific notation e.g. 1e-4.",
        default=None,
    )
    parser.add_argument("--loss", help="Should only be used to employ alternative Loss Function", default=None)
    parser.add_argument("--mom", help="Should only be used to employ alternative Momentum.", default=None)

    # Only touch the following if you know what you're doing #
    parser.add_argument(
        "--continue_train", help="continue training a previously saved checkpoint. ", action="store_true", default=False
    )
    # parser.add_argument("--chk", help="used to specify checkpoint to continue training from "
    #                    "when --continue_train is supplied. "
    #                    "The default is the latest model.", default='latest')
    # parser.add_argument("--threads", help="number of threads/processes", default=2)

    args = parser.parse_args()

    task = maybe_get_task_from_task_id(args.task)
    model = args.m
    dimensions = args.d
    trainer_name = args.tr
    plans = args.pl
    split_idx = args.f
    ensemble = args.ensemble
    fast_training = args.fast
    lr = args.lr
    loss = args.loss
    momentum = args.mom
    continue_training = args.continue_train
    # threads = args.threads
    # checkpoint = args.chk

    assert model in [
        "MedNeXt",
        "MultiResUNet",
        "UNet",
        "UNetR",
        "UXNet",
        "ResNet50",
    ], f"{model} is an invalid model name. This is case sensitive."

    if lr:
        assert "e" in lr, f"Learning Rate should be in scientific notation e.g. 1e-4, but is {lr}"

    if not ensemble:
        trainer = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "training", "trainers")],
            class_name=trainer_name,
            current_module="yucca.training.trainers",
        )
        trainer = trainer(
            model=model,
            model_dimensions=dimensions,
            task=task,
            split_idx=split_idx,
            plan_id=plans,
            starting_lr=lr,
            loss_fn=loss,
            momentum=momentum,
            continue_training=continue_training,
            fast_training=fast_training,
        )

        command_used = "yucca_train " + " ".join(f"-{k} {v}" for k, v in vars(args).items())
        trainer.set_train_command(command_used)

        trainer.run_training()
    if ensemble:
        print("Starting ensemble training. Model dimensions will automatically be set to 2D.")
        dimensions = "2D"
        views = ["X", "Y", "Z"]
        for view in views:
            trainer = recursive_find_python_class(
                split_idx=[join(yucca.__path__[0], "training", "trainers")],
                class_name=trainer_name,
                current_module="yucca.training.trainers",
            )
            plan_and_view = plans + view
            trainer = trainer(
                model=model,
                model_dimensions=dimensions,
                task=task,
                split_idx=split_idx,
                plan_id=plan_and_view,
                starting_lr=lr,
                loss_fn=loss,
                momentum=momentum,
                continue_training=continue_training,
                fast_training=fast_training,
            )

            command_used = "yucca_train " + " ".join(f"-{k} {v}" for k, v in vars(args).items())
            trainer.set_train_command(command_used)

            trainer.run_training()


if __name__ == "__main__":
    main()



import torch
from batchgenerators.utilities.file_and_folder_operations import join, isfile
from dataclasses import dataclass
from typing import Union
from yucca.training.configuration.configure_paths import PathConfig
from yucca.utils.dict import without_keys
import logging


@dataclass
class CkptConfig:
    """
    This class serves to preserve information between runs and should thus be extended to include
    information that must not be overwritten by each run.

    The LightningModule will call save_hyperparameters each time is it instantiated. This includes:
    starting training the first time, starting finetuning, continuing an interrupted training,
    starting inference. And each time the hyperparameters will be overwritten.

    ckpt_path will preserve the path to the original weights - this will by default also be saved
    when finetuning is started, but if finetuning is stopped and then started again the path is wiped.

    ckpt_seed ensures that we can continue from the same seed - this is mostly relevant for continuing
    interrupted trainings, but will also affect finetuning.

    ckpt_plans allows us to load plans from the checkpoint - this is used in inference where we may
    not be able to find a preprocessed folder with a plans.json.

    ckpt_wandb_id allows us to group runs by their "parent" wandb_id - this is useful for visualizing
    groups of finetuning runs by their source model/run.
    NB: THIS WILL NOT BE USED TO RESUME LOGGER STATES. Resuming loggers is currently not supported.
        for active discussion see: https://github.com/Lightning-AI/pytorch-lightning/issues/5342
        for workaround see: https://github.com/Lightning-AI/pytorch-lightning/issues/13524
    """

    base_experiment: Union[str, None]
    ckpt_path: Union[str, None]
    ckpt_patch_size: Union[str, None]
    ckpt_seed: Union[int, None]
    ckpt_plans: Union[dict, None]
    ckpt_version_dir: Union[str, None]
    ckpt_wandb_id: Union[str, None]

    def lm_hparams(self, without: [] = []):
        hparams = {
            "base_experiment": self.base_experiment,
            "ckpt_path": self.ckpt_path,
            "ckpt_patch_size": self.ckpt_patch_size,
            "ckpt_seed": self.ckpt_seed,
            "ckpt_plans": self.ckpt_plans,
            "ckpt_version_dir": self.ckpt_version_dir,
            "ckpt_wandb_id": self.ckpt_wandb_id,
        }
        return without_keys(hparams, without)


def get_checkpoint_config(
    continue_from_most_recent: bool, current_experiment: str, path_config: PathConfig, ckpt_path: Union[str, None] = None
):
    ckpt_path = find_checkpoint_path(
        ckpt_path=ckpt_path,
        continue_from_most_recent=continue_from_most_recent,
        version=path_config.version,
        version_dir=path_config.version_dir,
    )

    # If we did not find a checkpoint we just return an empty ckpt_config
    if ckpt_path is None:
        return CkptConfig(
            base_experiment=current_experiment,
            ckpt_path=None,
            ckpt_patch_size=None,
            ckpt_seed=None,
            ckpt_plans=None,
            ckpt_version_dir=None,
            ckpt_wandb_id=None,
        )

    return get_checkpoint_config_from_ckpt(ckpt_path)


def get_checkpoint_config_from_ckpt(ckpt_path: str):
    checkpoint = torch.load(ckpt_path, map_location="cpu")["hyper_parameters"]["config"]
    base_experiment, plans, patch_size, seed, version_dir, wandb_id = get_checkpoint_params(checkpoint)

    return CkptConfig(
        base_experiment=base_experiment,
        ckpt_path=ckpt_path,
        ckpt_patch_size=patch_size,
        ckpt_seed=seed,
        ckpt_plans=plans,
        ckpt_version_dir=version_dir,
        ckpt_wandb_id=wandb_id,
    )


def find_checkpoint_path(ckpt_path: Union[str, None], continue_from_most_recent: bool, version: int, version_dir: str):
    if ckpt_path:
        assert isfile(ckpt_path)
        logging.info(f"Using ckpt file: {ckpt_path}")
        return ckpt_path
    elif version is not None and continue_from_most_recent and isfile(join(version_dir, "checkpoints", "last.ckpt")):
        logging.info("Using last checkpoint and continuing training")
        return join(version_dir, "checkpoints", "last.ckpt")
    else:
        return None


def get_checkpoint_params(checkpoint: dict):
    base_experiment = checkpoint.get("base_experiment") if checkpoint.get("base_experiment") != "null" else None
    plans = checkpoint.get("plans") if checkpoint.get("plans") != "null" else None
    patch_size = checkpoint.get("patch_size") if checkpoint.get("patch_size") != "null" else None
    seed = checkpoint.get("seed") if checkpoint.get("seed") != "null" else None
    version_dir = checkpoint.get("version_dir") if checkpoint.get("version_dir") != "null" else None
    wandb_id = checkpoint.get("wandb_id") if checkpoint.get("wandb_id") != "null" else None
    return base_experiment, plans, patch_size, seed, version_dir, wandb_id



"""
Negative log likelihood loss wrapper
Makes loss function robust to target tensors with a channel dimension
"""

import torch
from torch import nn, Tensor


class NLL(nn.NLLLoss):
    def __init__(self, log=False):
        super().__init__()
        self.log = log

    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        if self.log:
            return super().forward(torch.log(input), target)
        return super().forward(input, target)



import numpy as np
import os
import torch
import wandb
import sys
import random
import yucca
from time import localtime, strftime, time, mktime
from abc import abstractmethod
from collections import OrderedDict
from torch import nn, autocast
from batchgenerators.utilities.file_and_folder_operations import (
    save_json,
    join,
    load_json,
    load_pickle,
    maybe_mkdir_p,
    isfile,
    subfiles,
)
from batchgenerators.dataloading.nondet_multi_threaded_augmenter import (
    NonDetMultiThreadedAugmenter,
)
from batchgenerators.transforms.abstract_transforms import Compose
from yucca.paths import yucca_preprocessed_data, yucca_models
from yucca.training.data_loading.YuccaLoader import YuccaLoader
from yucca.preprocessing.YuccaPreprocessor import YuccaPreprocessor
from yucca.image_processing.matrix_ops import get_max_rotated_size
from yucca.network_architectures.utils.model_memory_estimation import (
    find_optimal_tensor_dims,
)
from yucca.utils.files_and_folders import recursive_find_python_class
from yucca.utils.saving import save_prediction_from_logits
from yucca.utils.torch_utils import maybe_to_gpu
from yucca.image_processing.transforms.BiasField import BiasField
from yucca.image_processing.transforms.Blur import Blur
from yucca.image_processing.transforms.CopyImageToSeg import CopyImageToSeg
from yucca.image_processing.transforms.cropping_and_padding import CropPad
from yucca.image_processing.transforms.formatting import NumpyToTorch
from yucca.image_processing.transforms.Gamma import Gamma
from yucca.image_processing.transforms.Ghosting import MotionGhosting
from yucca.image_processing.transforms.Masking import Masking
from yucca.image_processing.transforms.Mirror import Mirror
from yucca.image_processing.transforms.Noise import (
    AdditiveNoise,
    MultiplicativeNoise,
)
from yucca.image_processing.transforms.Ringing import GibbsRinging
from yucca.image_processing.transforms.sampling import DownsampleSegForDS
from yucca.image_processing.transforms.SimulateLowres import SimulateLowres
from yucca.image_processing.transforms.Spatial import Spatial


class base_manager(object):
    def __init__(self):
        self.train_command = None
        self.is_initialized = False
        self.log_file = None
        self.param_dict = OrderedDict()
        self.network = None
        self.final_epoch = 1000
        self.current_epoch = 0
        self.is_seeded = False
        self.save_every = 300
        self.p_force_foreground = 0.33
        self.deep_supervision = False
        self.max_vram = 12

        # Debugging
        self.get_timings = True

        # These can be changed during inference
        self.threads_for_inference = 2
        self.threads_for_augmentation = 6
        self.sliding_window_overlap = 0.5
        self.preprocessor = None
        self.compression_level = 9  # 0-9 with 0 = min and 9 = max

        # These will always be set by the individual Trainer
        # Using arguments supplied by run_training
        self.train_batches_per_epoch = self.val_batches_per_epoch = self.batch_size_2D = self.batch_size_3D = self.folds = (
            self.model_dimensions
        ) = self.model_name = self.outpath = self.patch_size_2D = self.patch_size_3D = self.task = self.plan_id = self.name = (
            None
        )

        # These can be set by the individual Trainer
        # Using optional arguments supplied by run_training
        self.starting_lr = self.grad_scaler = self.loss_fn = self.loss_fn_kwargs = self.lr_scheduler = (
            self.lr_scheduler_kwargs
        ) = self.momentum = self.optim = self.optim_kwargs = self.random_seed = self.fast_training = self.finetune = (
            self.fast_train_batches_per_epoch
        ) = self.fast_val_batches_per_epoch = self.fast_final_epoch = None

        # These will always be set by the plans file
        self.classes = self.nclasses = self.modalities = self.nmodalities = self.folder_with_preprocessed_data = self.plans = (
            self.plans_path
        ) = None

        # These will be used during training
        self.tr_losses = []
        self.val_losses = []
        self.best_val_loss = 99999

    @abstractmethod
    def comprehensive_eval(self):
        """
        implement in trainer subclass
        """
        pass

    @abstractmethod
    def initialize(self):
        """
        implement in trainer subclass
        """
        pass

    @abstractmethod
    def initialize_loss_optim_lr(self):
        """
        implement in trainer subclass
        """
        pass

    @abstractmethod
    def run_training(self):
        """
        implement this in subclasses
        """
        pass

    def setup_DA(self):
        # Define whether we crop before or after applying augmentations
        # Define if cropping is random or always centered
        self.EarlyCrop = True
        self.RandomCrop = True
        self.MaskImageForReconstruction = False

        # Label/segmentation transforms
        self.SkipSeg = False
        self.SegDtype = int
        self.CopyImageToSeg = False

        self.AdditiveNoise_p_per_sample = 0.0
        self.AdditiveNoise_mean = (0.0, 0.0)
        self.AdditiveNoise_sigma = (1e-3, 1e-4)

        self.BiasField_p_per_sample = 0.0

        self.Blurring_p_per_sample = 0.0
        self.Blurring_sigma = (0.0, 1.0)
        self.Blurring_p_per_channel = 0.5

        self.ElasticDeform_p_per_sample = 0.0
        self.ElasticDeform_alpha = (200, 600)
        self.ElasticDeform_sigma = (20, 30)

        self.Gamma_p_per_sample = 0.0
        self.Gamma_p_invert_image = 0.05
        self.Gamma_range = (0.5, 2.0)

        self.GibbsRinging_p_per_sample = 0.0
        self.GibbsRinging_cutFreq = (96, 129)
        self.GibbsRinging_axes = (0, 3)

        self.Mirror_p_per_sample = 0.0
        self.Mirror_p_per_axis = 0.33
        self.Mirror_axes = (0, 1, 2)

        self.MotionGhosting_p_per_sample = 0.0
        self.MotionGhosting_alpha = (0.85, 0.95)
        self.MotionGhosting_numReps = (2, 11)
        self.MotionGhosting_axes = (0, 3)

        self.MultiplicativeNoise_p_per_sample = 0.0
        self.MultiplicativeNoise_mean = (0, 0)
        self.MultiplicativeNoise_sigma = (1e-3, 1e-4)

        self.Rotation_p_per_sample = 0.0
        self.Rotation_p_per_axis = 0.66
        self.Rotation_x = (-30.0, 30.0)
        self.Rotation_y = (-30.0, 30.0)
        self.Rotation_z = (-30.0, 30.0)

        self.Scale_p_per_sample = 0.0
        self.Scale_factor = (0.9, 1.1)

        self.SimulateLowres_p_per_sample = 0.0
        self.SimulateLowres_p_per_channel = 0.5
        self.SimulateLowres_p_per_axis = 0.33
        self.SimulateLowres_zoom_range = (0.5, 1.0)

        if self.model_dimensions == "2D":
            self.GibbsRinging_axes = (0, 2)
            self.Mirror_axes = (0, 1)
            self.MotionGhosting_axes = (0, 2)
            self.Rotation_y = (-0.0, 0.0)
            self.Rotation_z = (-0.0, 0.0)

    @abstractmethod
    def split_data(self):
        """
        implement in trainer subclass
        """
        pass

    def epoch_finish(self):
        self.epoch_end_time = time()
        self.log(f'{"Current Epoch:":20}', self.current_epoch)
        self.log(f'{"Training Loss:":20}', round(self.tr_losses[-1], 3))
        self.log(f'{"Validation Loss:":20}', round(self.val_losses[-1], 3))
        wandb.log(
            {
                "Training Loss": round(self.tr_losses[-1], 3),
                "Validation Loss": round(self.val_losses[-1], 3),
            },
            commit=False,
        )
        # self.log(f'{"Validation Dice:":15}', )
        # self.log(f'{":":15}', )

        if self.epoch_eval_dict:
            epoch_eval_dict_per_label = {}
            for key in self.epoch_eval_dict:
                self.epoch_eval_dict[key] = np.round(np.nanmean(self.epoch_eval_dict[key], 0), 4)
                self.log(f"{key:20}", self.epoch_eval_dict[key])
                for i, val in enumerate(self.epoch_eval_dict[key]):
                    epoch_eval_dict_per_label[f"{key}_{i+1}"] = val
            wandb.log(epoch_eval_dict_per_label, commit=False)

        if round(self.val_losses[-1], 3) < self.best_val_loss:
            self.best_val_loss = round(self.val_losses[-1], 3)
            wandb.log({"Best Validation Loss": self.best_val_loss}, commit=False)
            self.save_checkpoint("checkpoint_best.model")

        if self.current_epoch % self.save_every == 0:
            self.save_checkpoint(f"checkpoint_{self.current_epoch}.model")
        elif self.current_epoch % 50 == 0:
            self.save_checkpoint(f"checkpoint_latest.model")

        wandb.log({"Learning Rate": self.optim.param_groups[0]["lr"]}, commit=False)
        self.log(f'{"Learning Rate:":20}', self.optim.param_groups[0]["lr"])

        self.epoch_time = self.epoch_end_time - self.epoch_start_time
        self.log(f'{"Time elapsed:":20}', self.epoch_time)
        # This is THE log call that will update the "step" counter in wandb.
        # All others wandb.log calls (should) have commit=False.
        wandb.log({"Time/epoch": self.epoch_end_time - self.epoch_start_time})

        self.current_epoch += 1
        sys.stdout.flush()

    def epoch_start(self):
        self.epoch_tr_loss = []
        self.epoch_val_loss = []
        self.epoch_eval_dict = None
        self.epoch_start_time = time()
        self.log("\n", time=False, also_print=True)

    def initialize_network(self):
        self.network = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "network_architectures")],
            class_name=self.model_name,
            current_module="yucca.network_architectures",
        )

        if self.model_dimensions == "3D":
            conv = nn.Conv3d
            norm = nn.InstanceNorm3d
            dropout = nn.Dropout3d
        else:
            conv = nn.Conv2d
            norm = nn.InstanceNorm2d
            dropout = nn.Dropout2d

        self.network = self.network(
            input_channels=self.nmodalities,
            num_classes=self.nclasses,
            conv_op=conv,
            norm_op=norm,
            dropout_op=dropout,
            patch_size=self.patch_size,
            deep_supervision=self.deep_supervision,
        )

        if torch.cuda.is_available():
            self.network.cuda()

    def initialize_preprocessor_for_inference(self):
        preprocessor_class = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "preprocessing")],
            class_name=self.plans["preprocessor"],
            current_module="yucca.preprocessing",
        )
        assert preprocessor_class, f"searching for {self.plans['preprocessor']}" f"but found: {preprocessor_class}"
        assert issubclass(preprocessor_class, YuccaPreprocessor), "Preprocessor is not a subclass " "of YuccaPreprocesor"
        print(f"{'Using preprocessor: ':25} {preprocessor_class}")
        self.preprocessor = preprocessor_class(self.plans_path)

    def initialize_wandb(self):
        wandb.init(
            project="Yucca",
            group=self.task,
            dir=self.outpath,
            name=join(os.path.split(self.outpath)[-1] + "_" + strftime("%Y_%m_%d_%H_%M_%S", self.initial_time_obj)),
            config=self.param_dict,
        )

    def load_checkpoint(self, checkpoint, train=False):
        """
        If we are loading a checkpoint to continue training or finetune etc., then we need to
        initialize() to have outpaths, training data, logging and data augmentation

        If we are loading a checkpoint for inference we just need to initialize the network
        and then apply the pretrained weights
        """

        self.load_plans_from_path(checkpoint + ".json")

        chk = torch.load(checkpoint, map_location=torch.device("cpu"))

        if self.finetune:
            # Filter out layers from old model if they dont match sizes of the new model
            # e.g. when using a model pretrained on 1 label for a multi-label problem.
            # this will discard the weights final layer
            state_dict = {
                k: v
                for k, v in chk["model_state_dict"].items()
                if k in self.network.state_dict().keys() and v.shape == self.network.state_dict()[k].shape
            }
            # Then make sure optimizer layers are appropriate, otherwise re-initialize them
            # with appropriate dimensions
            for idx, layer in enumerate(self.network.parameters()):
                if layer.shape != chk["optimizer_state_dict"]["state"][idx]["momentum_buffer"]:
                    chk["optimizer_state_dict"]["state"][idx]["momentum_buffer"] = torch.zeros(layer.shape)
        else:
            state_dict = chk["model_state_dict"]

        if "patch_size" in chk:
            self.patch_size = chk["patch_size"]
        else:
            print(
                "patch_size not found in checkpoint. Generating a new one now, better hope "
                "it's the same as the training patch_size."
            )
            self.set_batch_and_patch_sizes()

        if train:
            self.optim.load_state_dict(chk["optimizer_state_dict"])
            self.current_epoch = chk["epoch"]
            self.tr_losses = chk["tr_losses"]
            self.val_losses = chk["val_losses"]
            self.best_val_loss = chk["best"]
            if self.grad_scaler is not None:
                self.grad_scaler.load_state_dict(chk["grad_scaler_state_dict"])
            self.log("Continuing training from checkpoint:", checkpoint, time=False)
            if self.finetune:
                self.best_val_loss = 99999
                self.final_epoch = 500
                self.current_epoch = 0
                for param in self.optim.param_groups:
                    param["lr"] = self._DEFAULT_STARTING_LR * 0.1
        else:
            self.initialize_network()

        self.network.load_state_dict(state_dict, strict=False)

        if torch.cuda.is_available():
            self.network.cuda()

    def load_data(self):
        assert len(self.patch_size) in [2, 3], "Patch Size should be (x, y, z) or (x, y)" f" but is: {self.patch_size}"

        if not self.is_seeded:
            self.set_random_seeds()

        self.folder_with_preprocessed_data = join(yucca_preprocessed_data, self.task, self.plan_id)
        self.splits_file = join(yucca_preprocessed_data, self.task, "splits.pkl")
        self.log(f'{"data folder:":20}', self.folder_with_preprocessed_data, time=False)

        if not isfile(self.splits_file):
            self.split_data()

        self.splits = load_pickle(self.splits_file)[self.folds]

        self.train_samples = [join(self.folder_with_preprocessed_data, sample) for sample in self.splits["train"]]
        self.val_samples = [join(self.folder_with_preprocessed_data, sample) for sample in self.splits["val"]]
        """Here we want to calculate a larger-than-final patch size
        to avoid cropping out parts that would be rotated into inclusion
        and to avoid large interpolation artefacts near the borders of our final patch
        For this we use the hypotenuse of the 2 largest dimensions"""
        self.initial_patch_size = get_max_rotated_size(self.patch_size)

        self.log(f'{"patch size pre-aug:":20}', self.initial_patch_size, time=False)
        self.log(f'{"patch size final:":20}', self.patch_size, time=False)
        self.log(f'{"batch size:":20}', self.batch_size, time=False)
        self.log(f'{"classes":20}', self.classes, time=False)
        self.log(f'{"modalities:":20}', self.modalities, time=False)
        self.log(f'{"fold:":20}', self.folds, time=False)
        self.log(f'{"train batches/epoch:":20}', self.train_batches_per_epoch, time=False)
        self.log(f'{"val batches/epoch:":20}', self.val_batches_per_epoch, time=False)

        self.tr_loader = YuccaLoader(
            self.train_samples,
            self.batch_size,
            self.initial_patch_size,
            p_oversample_foreground=self.p_force_foreground,
        )
        self.val_loader = YuccaLoader(
            self.val_samples,
            self.batch_size,
            self.patch_size,
            p_oversample_foreground=self.p_force_foreground,
        )

        self.get_data_generators()

    def get_data_generators(self):
        train_transforms = []
        val_transforms = []

        # Augmentations
        train_transforms.append(
            Spatial(
                crop=self.EarlyCrop,
                random_crop=self.RandomCrop,
                patch_size=self.patch_size,
                p_deform_per_sample=self.ElasticDeform_p_per_sample,
                deform_sigma=self.ElasticDeform_sigma,
                deform_alpha=self.ElasticDeform_alpha,
                p_rot_per_sample=self.Rotation_p_per_sample,
                p_rot_per_axis=self.Rotation_p_per_axis,
                x_rot_in_degrees=self.Rotation_x,
                y_rot_in_degrees=self.Rotation_y,
                z_rot_in_degrees=self.Rotation_z,
                p_scale_per_sample=self.Scale_p_per_sample,
                scale_factor=self.Scale_factor,
                skip_seg=self.SkipSeg,
            )
        )

        train_transforms.append(
            AdditiveNoise(
                p_per_sample=self.AdditiveNoise_p_per_sample,
                mean=self.AdditiveNoise_mean,
                sigma=self.AdditiveNoise_sigma,
            )
        )

        train_transforms.append(
            Blur(
                p_per_sample=self.Blurring_p_per_sample,
                p_per_channel=self.Blurring_p_per_channel,
                sigma=self.Blurring_sigma,
            )
        )

        train_transforms.append(
            MultiplicativeNoise(
                p_per_sample=self.MultiplicativeNoise_p_per_sample,
                mean=self.MultiplicativeNoise_mean,
                sigma=self.MultiplicativeNoise_sigma,
            )
        )

        train_transforms.append(
            MotionGhosting(
                p_per_sample=self.MotionGhosting_p_per_sample,
                alpha=self.MotionGhosting_alpha,
                numReps=self.MotionGhosting_numReps,
                axes=self.MotionGhosting_axes,
            )
        )

        train_transforms.append(
            GibbsRinging(
                p_per_sample=self.GibbsRinging_p_per_sample,
                cutFreq=self.GibbsRinging_cutFreq,
                axes=self.GibbsRinging_axes,
            )
        )

        train_transforms.append(
            SimulateLowres(
                p_per_sample=self.SimulateLowres_p_per_sample,
                p_per_channel=self.SimulateLowres_p_per_channel,
                p_per_axis=self.SimulateLowres_p_per_axis,
                zoom_range=self.SimulateLowres_zoom_range,
            )
        )

        train_transforms.append(BiasField(p_per_sample=self.BiasField_p_per_sample))

        train_transforms.append(
            Gamma(
                p_per_sample=self.Gamma_p_per_sample,
                p_invert_image=self.Gamma_p_invert_image,
                gamma_range=self.Gamma_range,
            )
        )

        train_transforms.append(
            Mirror(
                p_per_sample=self.Mirror_p_per_sample,
                axes=self.Mirror_axes,
                p_mirror_per_axis=self.Mirror_p_per_axis,
                skip_seg=self.SkipSeg,
            )
        )

        if not self.EarlyCrop:
            train_transforms.append(CropPad(patch_size=self.patch_size, random_crop=self.RandomCrop))

        if self.deep_supervision:
            train_transforms.append(DownsampleSegForDS())

        if self.CopyImageToSeg:
            train_transforms.append(CopyImageToSeg())

        if self.MaskImageForReconstruction:
            train_transforms.append(Masking())

        train_transforms.append(NumpyToTorch(seg_dtype=self.SegDtype))
        train_transforms = Compose(train_transforms)

        self.tr_gen = NonDetMultiThreadedAugmenter(
            self.tr_loader,
            train_transforms,
            self.threads_for_augmentation,
            2,
            pin_memory=True,
        )

        # Validation Transforms
        if self.deep_supervision:
            val_transforms.append(DownsampleSegForDS())

        if self.CopyImageToSeg:
            val_transforms.append(CopyImageToSeg())

        if self.MaskImageForReconstruction:
            val_transforms.append(Masking())

        val_transforms.append(NumpyToTorch(seg_dtype=self.SegDtype))
        val_transforms = Compose(val_transforms)
        self.val_gen = NonDetMultiThreadedAugmenter(
            self.val_loader,
            val_transforms,
            self.threads_for_augmentation // 2,
            2,
            pin_memory=True,
        )

    def load_plans_from_path(self, path):
        # if plans already exist, it is because we are finetuning.
        # In this case we want to combine plans from two tasks, keeping dataset properties
        # from the already loaded plans, and the rest (such as normalization, tranposition etc.,)
        # from the one we are loading afterwards.
        self.plans_path = path

        if self.plans:
            properties = self.plans["dataset_properties"]
            self.plans = load_json(self.plans_path)
            self.plans["dataset_properties"] = properties
        else:
            self.plans = load_json(self.plans_path)

        self.classes = self.plans["dataset_properties"]["classes"]
        self.nclasses = len(self.classes)

        self.modalities = self.plans["dataset_properties"]["modalities"]
        self.nmodalities = len(self.modalities)

        self.plans["trainer_class"] = self.__class__.__name__

        if self.model_dimensions is None:
            self.model_dimensions = self.plans["suggested_dimensionality"]

    def log(self, *args, time=True, also_print=True):
        if not self.log_file:
            self.initial_time_obj = localtime()
            self.log_file = join(
                self.outpath,
                "log_" + strftime("%Y_%m_%d_%H_%M_%S", self.initial_time_obj) + ".txt",
            )
            with open(self.log_file, "w") as f:
                f.write("Starting model training")
                print("Starting model training \n" f'{"log file:":20} {self.log_file} \n')
                f.write("\n")
                f.write(f'{"log file:":20} {self.log_file}')
                f.write("\n")
        t = strftime("%Y_%m_%d_%H_%M_%S", localtime())
        with open(self.log_file, "a+") as f:
            if time:
                f.write(t)
                f.write(" ")
            for arg in args:
                f.write(str(arg))
                f.write(" ")
            f.write("\n")
        if also_print:
            print(" ".join([str(arg) for arg in args]))

    def predict_folder(
        self,
        input_folder,
        output_folder,
        not_strict=True,
        save_softmax=False,
        overwrite=False,
        do_tta=False,
    ):
        self.initialize_preprocessor_for_inference()

        files = subfiles(input_folder, suffix=".nii.gz", join=False)

        if not not_strict:
            # If strict we enforce modality encoding.
            # This means files must be encoded as the model expects.
            # e.g. "_000" or "_001" etc., for T1 and T2 scans.
            # This allows us to handle different modalities
            # of the same subject, rather than treating them as individual cases.
            expected_modalities = self.plans["dataset_properties"]["modalities"]
            subject_ids, counts = np.unique([i[: -len("_000.nii.gz")] for i in files], return_counts=True)
            assert all(counts == len(expected_modalities)), "Aborting. Modalities are missing for some samples"
        else:
            subject_ids = np.unique([i[: -len(".nii.gz")] for i in files])

        all_cases = []
        all_outpaths = []
        for subject_id in subject_ids:
            case = [
                impath
                for impath in subfiles(input_folder, suffix=".nii.gz")
                if os.path.split(impath)[-1][: -len("_000.nii.gz")] == subject_id
            ]
            outpath = join(output_folder, subject_id)
            all_cases.append([case, outpath, save_softmax, overwrite, do_tta])
            all_outpaths.append(outpath)

        n_already_predicted = len(subfiles(output_folder, suffix=".nii.gz"))

        print(
            f"\n"
            f"STARTING PREDICTION \n"
            f"{'Cases already predicted: ':25} {n_already_predicted} \n"
            f"{'Cases NOT predicted: ':25} {len(all_outpaths) - n_already_predicted} \n"
            f"{'Overwrite predictions: ':25} {overwrite} \n"
        )

        for case_info in all_cases:
            self.preprocess_predict_test_case(*case_info)

    def preprocess_predict_test_case(self, case, outpath, save_softmax=False, overwrite=False, do_tta=False):
        """
        Load the preprocessor if not already loaded (e.g. in single prediction cases)
        Preprocessor will apply the same type of preprocessing that was applied to training data,
        and transpose to target view if necessary as well as handling data formats (torch GPU/CPU)
        Finally we detach, move tensor to GPU and convert to np.ndarray before transposing it
        to the original view.
        """

        # If overwrite = True, we don't care if file already exists
        # If overwrite = False, check if files already exist. If already exists we skip.

        if not overwrite:
            if isfile(outpath + ".nii.gz"):
                if not save_softmax:
                    return
                if save_softmax:
                    if isfile(outpath + ".npz"):
                        return

        if not self.preprocessor:
            self.initialize_preprocessor_for_inference()

        print(f"{'Predicting case: ':25} {case}")
        case, image_properties = self.preprocessor.preprocess_case_for_inference(case, self.patch_size)

        logits = (
            self.network.predict(
                mode=self.model_dimensions,
                data=case,
                patch_size=self.patch_size,
                overlap=self.sliding_window_overlap,
                mirror=do_tta,
            )
            .detach()
            .cpu()
            .numpy()
        )

        logits = self.preprocessor.reverse_preprocessing(logits, image_properties)

        print(f"{'Saving as: ':25} {outpath}")
        save_prediction_from_logits(
            logits,
            outpath=outpath,
            properties=image_properties,
            save_softmax=save_softmax,
            compression=self.compression_level,
        )
        print("\n")

    def run_batch(self, batch, train=True, comprehensive_eval=False):
        self.optim.zero_grad()

        image = batch["image"]
        seg = batch["seg"]

        image = maybe_to_gpu(image)
        seg = maybe_to_gpu(seg)

        with autocast(device_type="cuda", dtype=torch.float16, enabled=True):
            pred = self.network(image)
            del image
            loss = self.loss_fn(pred, seg)

        if comprehensive_eval:
            # Add additional_eval here to retrieve e.g. Dice score, #TP, #FP, etc.
            self.comprehensive_eval(pred, seg)
        del seg

        if train:
            self.grad_scaler.scale(loss).backward()
            self.grad_scaler.unscale_(self.optim)
            # The max norm here is a mythical setting with limited documentation/experiments. Just trust.
            torch.nn.utils.clip_grad_norm_(self.network.parameters(), max_norm=12)
            self.grad_scaler.step(self.optim)
            self.grad_scaler.update()

        return loss.detach().cpu().numpy()

    def save_checkpoint(self, fname):
        self.log(f'{"Saving model:":20}', fname)
        model_dict = {
            "epoch": self.current_epoch,
            "model_state_dict": self.network.state_dict(),
            "optimizer_state_dict": self.optim.state_dict(),
            "scheduler_state_dict": self.lr_scheduler.state_dict(),
            "tr_losses": self.tr_losses,
            "val_losses": self.val_losses,
            "best": self.best_val_loss,
            "plans": self.plans_path,
            "patch_size": self.patch_size,
        }
        if self.grad_scaler is not None:
            model_dict["grad_scaler_state_dict"] = self.grad_scaler.state_dict()
        torch.save(model_dict, join(self.outpath, fname))
        save_json(self.plans, join(self.outpath, fname + ".json"))

    def save_parameter_json(self):
        # Model settings
        self.param_dict["network"] = self.model_name
        self.param_dict["model dimensions"] = self.model_dimensions

        # Data
        self.param_dict["patch_size"] = self.patch_size
        self.param_dict["plans id"] = self.plan_id
        self.param_dict["folds"] = self.folds
        self.param_dict["classes"] = self.classes
        self.param_dict["modalities"] = self.modalities
        self.param_dict["batch size"] = self.batch_size
        self.param_dict["max epochs"] = self.final_epoch
        self.param_dict["train batches per epoch"] = self.train_batches_per_epoch
        self.param_dict["val batches per epoch"] = self.val_batches_per_epoch
        self.param_dict["foreground oversampling"] = self.p_force_foreground
        self.param_dict["train_samples"] = self.train_samples
        self.param_dict["val_samples"] = self.val_samples

        # Parameters
        self.param_dict["loss function"] = self.loss_fn.__class__.__name__
        self.param_dict["loss kwargs"] = self.loss_fn_kwargs
        self.param_dict["optimizer"] = self.optim.__class__.__name__
        self.param_dict["optimizer kwargs"] = self.optim_kwargs
        self.param_dict["momentum"] = self.momentum
        self.param_dict["starting lr"] = self.starting_lr
        # self.param_dict['parameters for augmentation'] = selfmentation_parameters

        if self.lr_scheduler:
            self.param_dict["lr scheduler"] = self.lr_scheduler.__class__.__name__
            self.param_dict["lr scheduler kwargs"] = self.lr_scheduler_kwargs
        if self.grad_scaler:
            self.grad_scaler.__class__.__name__

        # Paths
        self.param_dict["output folder"] = self.outpath
        self.param_dict["log file"] = self.log_file
        self.param_dict["folder with preprocessed files"] = self.folder_with_preprocessed_data

        # Experiment
        self.param_dict["task"] = self.task
        self.param_dict["name"] = self.name
        self.param_dict["random seed"] = self.random_seed
        self.param_dict["command used"] = self.train_command
        save_json(self.param_dict, join(self.outpath, "parameters.json"), sort_keys=False)

    def set_batch_and_patch_sizes(self):
        self.batch_size, self.patch_size = find_optimal_tensor_dims(
            dimensionality=self.model_dimensions,
            num_classes=self.nclasses,
            modalities=self.nmodalities,
            model_name=self.model_name,
            max_patch_size=self.plans["new_mean_size"],
            max_memory_usage_in_gb=self.max_vram,
        )

    def set_train_length(self):
        if self.fast_training:
            self.train_batches_per_epoch = self.fast_train_batches_per_epoch
            self.val_batches_per_epoch = self.fast_val_batches_per_epoch
            self.final_epoch = self.fast_final_epoch

    def set_train_command(self, command):
        self.train_command = command

    def set_outpath(self):
        if not self.outpath:
            if self.starting_lr:
                self.name += "_" + str(self.starting_lr).replace("-", "")

            if self.loss_fn:
                self.name += "_" + str(self.loss_fn)

            if self.momentum:
                self.name += "_" + "m" + str(self.momentum).replace("0.", "")

            if self.fast_training:
                self.name += "_Fast"

            if self.finetune:
                self.name += "_FT"

            self.name += "__" + self.plan_id

            self.outpath = join(
                yucca_models,
                self.task,
                self.model_name,
                self.model_dimensions,
                self.name,
                str(self.folds),
            )

            maybe_mkdir_p(self.outpath)
            self.log(f'{"outpath:":20}', self.outpath, time=False)
        else:
            print("outpath already defined")

    def set_random_seeds(self, seed=None):
        if not self.is_seeded:
            if seed is not None:
                self.random_seed = int(seed)
                self.log("Using the random seed: ", self.random_seed, time=False)
            else:
                self.random_seed = int(mktime(self.initial_time_obj))
                self.log(
                    "No random seed using initial time object: ",
                    self.random_seed,
                    time=False,
                )
            torch.manual_seed(self.random_seed)
            np.random.seed(self.random_seed)
            random.seed(self.random_seed)
            self.is_seeded = True
        else:
            print("Random seed already set")



import nibabel.orientations as nio
import numpy as np
import nibabel as nib


def get_nib_spacing(nib_image: nib.Nifti1Image) -> np.ndarray:
    return np.array(nib_image.header.get_zooms())


def get_nib_orientation(nib_image: nib.Nifti1Image) -> str:
    affine = nib_image.affine
    return "".join(nio.aff2axcodes(affine))


def reorient_nib_image(nib_image, original_orientation: str, target_orientation: str) -> np.ndarray:
    # The reason we don't use the affine information to get original_orientation is that it can be
    # incorrect. Therefore it can be manually specified. In the cases where header can be trusted,
    # Just use get_nib_orientation to get the original_orientation.
    if original_orientation == target_orientation:
        return nib_image
    start = nio.axcodes2ornt(original_orientation)
    end = nio.axcodes2ornt(target_orientation)
    orientation = nio.ornt_transform(start, end)
    return nib_image.as_reoriented(orientation)



from torchmetrics.classification import MulticlassF1Score
from torch import Tensor


class F1(MulticlassF1Score):
    def forward(self, input: Tensor, target: Tensor) -> Tensor:
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return super().forward(input, target)



import argparse
import yucca
from yucca.utils.task_ids import maybe_get_task_from_task_id
from yucca.paths import yucca_raw_data, yucca_results, yucca_models
from yucca.evaluation.YuccaEvaluator import YuccaEvaluator
from yucca.training.managers.YuccaManager import YuccaManager
from yucca.utils.files_and_folders import recursive_find_python_class
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    isfile,
    maybe_mkdir_p,
    isdir,
    subdirs,
)
from warnings import filterwarnings

filterwarnings("ignore")


def main():
    from warnings import filterwarnings

    filterwarnings("ignore")
    parser = argparse.ArgumentParser()
    # Required Arguments
    parser.add_argument(
        "-s",
        help="Name of the source task i.e. what the model is trained on. " "Should be of format: TaskXXX_MYTASK",
        required=True,
    )
    parser.add_argument(
        "-t",
        help="Name of the target task i.e. the data to be predicted. " "Should be of format: TaskXXX_MYTASK",
        required=True,
    )

    # Optionals (frequently changed)
    parser.add_argument(
        "-chk",
        "--checkpoint",
        help="Checkpoint to use for inference. Defaults to model_best.",
        default="best",
    )
    parser.add_argument("-d", "--dimensions", help="2D or 3D model. Defaults to 3D.", default="3D")
    parser.add_argument("-m", "--model", help="Model Architecture. Defaults to UNet.", default="UNet")
    parser.add_argument(
        "-man",
        "--manager",
        help="Full name of Trainer Class. \n" "e.g. 'YuccaTrainer_DCE' or 'YuccaTrainer'. Defaults to YuccaTrainer.",
        default="YuccaManager",
    )
    parser.add_argument("-pl", "--planner", help="Planner. Defaults to YuccaPlanner", default="YuccaPlanner")
    parser.add_argument("--split_idx", type=int, help="idx of splits to use for training.", default=0)
    parser.add_argument(
        "--split_data_method", help="Specify splitting method. Either kfold, simple_train_val_split", default="kfold"
    )
    parser.add_argument(
        "--split_data_param",
        help="Specify the parameter for the selected split method. For KFold use an int, for simple_split use a float between 0.0-1.0.",
        default=5,
    )
    parser.add_argument(
        "--task_type",
        default="segmentation",
        type=str,
        required=False,
        help="Defaults to segmentation. Set to 'classification' for classification tasks.",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="Version to use for inference. Defaults to the newest version.",
        default=None,
    )

    # Optionals (occasionally changed)
    parser.add_argument(
        "--experiment",
        help="A name for the experiment being performed, with no spaces.",
        default="default",
    )
    parser.add_argument(
        "--disable_tta",
        help="Used to disable test-time augmentations (mirroring)",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--no_eval",
        help="Disable evaluation and creation of metrics file (result.json)",
        default=False,
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "--no_wandb",
        help="Disable logging of evaluation results to wandb",
        default=False,
        action="store_true",
        required=False,
    )
    parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        required=False,
        help="Overwrite existing predictions",
    )
    parser.add_argument(
        "--predict_train",
        default=False,
        action="store_true",
        required=False,
        help="Predict on the training set. Useful for debugging.",
    )
    parser.add_argument(
        "--profile",
        help="Used to enable inference profiling",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--save_softmax",
        default=False,
        action="store_true",
        required=False,
        help="Save softmax outputs. Required for softmax fusion.",
    )

    args = parser.parse_args()

    # Required
    source_task = maybe_get_task_from_task_id(args.s)
    target_task = maybe_get_task_from_task_id(args.t)

    # Optionals (frequently changed)
    checkpoint = args.checkpoint
    dimensions = args.dimensions
    manager_name = args.manager
    model = args.model
    planner = args.planner
    split_idx = args.split_idx
    split_data_method = args.split_data_method
    split_data_param = args.split_data_param
    task_type = args.task_type
    version = args.version

    # Optionals (occasionally changed)
    experiment = args.experiment
    disable_tta = args.disable_tta
    no_eval = args.no_eval
    overwrite = args.overwrite
    predict_train = args.predict_train
    profile = args.profile
    save_softmax = args.save_softmax
    use_wandb = not args.no_wandb

    path_to_versions = join(
        yucca_models,
        source_task,
        model + "__" + dimensions,
        manager_name + "__" + planner,
        experiment,
        f"{split_data_method}_{split_data_param}_fold_{split_idx}",
    )
    if version is None:
        versions = [int(i.split("_")[-1]) for i in subdirs(path_to_versions, join=False)]
        version = str(max(versions))

    modelfile = join(
        yucca_models,
        source_task,
        model + "__" + dimensions,
        manager_name + "__" + planner,
        experiment,
        f"{split_data_method}_{split_data_param}_fold_{split_idx}",
        f"version_{version}",
        "checkpoints",
        checkpoint + ".ckpt",
    )

    assert isfile(modelfile), "Can't find .cpkt file with trained model weights. " f"Should be located at: {modelfile}"
    print(f"######################################################################## \n" f"{'Using model: ':25} {modelfile}")

    manager = recursive_find_python_class(
        folder=[join(yucca.__path__[0], "training", "managers")],
        class_name=manager_name,
        current_module="yucca.training.managers",
    )

    assert manager, f"searching for {manager_name} " f"but found: {manager}"
    assert issubclass(manager, (YuccaManager, YuccaManager)), "Trainer is not a subclass of YuccaTrainer."

    print(f"{'Using manager: ':25} {manager_name}")
    manager = manager(
        ckpt_path=modelfile,
        enable_logging=False,
        experiment=experiment,
        model_name=model,
        model_dimensions=dimensions,
        task=source_task,
        split_idx=split_idx,
        split_data_method=split_data_method,
        split_data_param=split_data_param,
        planner=planner,
        profile=profile,
    )

    # Setting up input paths and output paths
    inpath = join(yucca_raw_data, target_task, "imagesTs")
    ground_truth = join(yucca_raw_data, target_task, "labelsTs")

    outpath = join(
        yucca_results,
        target_task,
        source_task,
        model + "__" + dimensions,
        manager_name + "__" + planner,
        f"{split_data_method}_{split_data_param}_fold_{split_idx}",
        f"version_{version}",
        checkpoint,
    )

    if predict_train:
        inpath = join(yucca_raw_data, target_task, "imagesTr")
        ground_truth = join(yucca_raw_data, target_task, "labelsTr")
        outpath += "Tr"

    maybe_mkdir_p(outpath)

    manager.predict_folder(
        inpath,
        disable_tta,
        overwrite_predictions=overwrite,
        output_folder=outpath,
        save_softmax=save_softmax,
        # overwrite=overwrite, # Commented out until overwrite arg is added in manager.
    )

    if isdir(ground_truth) and not no_eval:
        evaluator = YuccaEvaluator(
            manager.model_module.num_classes,
            folder_with_predictions=outpath,
            folder_with_ground_truth=ground_truth,
            overwrite=overwrite,
            task_type=task_type,
            use_wandb=use_wandb,
        )
        evaluator.run()


if __name__ == "__main__":
    main()



import torch
import torch.nn.functional as F
from torch import nn, Tensor
from yucca.training.loss_and_optim.utils.LossTree import LossTree
from yucca.training.loss_and_optim.loss_functions.nnUNet_losses import DiceCE
from yucca.training.loss_and_optim.loss_functions.NLL import NLL


class HierarchicalLoss(nn.Module):
    def __init__(self, hierarchical_kwargs={}):
        super(HierarchicalLoss, self).__init__()
        self.root = LossTree(hierarchical_kwargs["rootdict"])  # root represents your hierarchy
        self.loss = NLL(log=True)

    def calculate_loss(self, running_loss, summed_probabilities, target):
        summed_probabilities /= torch.sum(summed_probabilities, axis=1).unsqueeze(1)
        level_loss = self.loss(summed_probabilities, target)
        return running_loss + level_loss

    def get_loss(self, running_loss, summed_probabilities, target, only_last=False):
        if only_last:
            if self.n_level + 1 == len(self.root.loss_layers):
                return self.calculate_loss(running_loss, summed_probabilities, target)
            else:
                return 0
        return self.calculate_loss(running_loss, summed_probabilities, target)

    def forward(self, logits: Tensor, target: Tensor) -> Tensor:
        probabilities = F.softmax(logits, dim=1)
        loss = 0
        precomputed_hierarchy_list = self.root.loss_layers
        for n_level, level_loss_list in enumerate(precomputed_hierarchy_list):
            self.n_level = n_level
            probabilities_tosum = probabilities.clone()
            summed_probabilities = probabilities_tosum
            for branch in level_loss_list:
                # Extract the relevant probabilities according to a branch in our hierarchy.
                branch_probs = torch.FloatTensor()
                branch_probs = branch_probs.to(logits.device)
                for channel in branch:
                    branch_probs = torch.cat((branch_probs, probabilities_tosum[:, channel].unsqueeze(1)), 1)

                # Sum these probabilities into a single slice; this is hierarchical inference.
                summed_tree_branch_slice = branch_probs.sum(1, keepdim=True)

                # Insert inferred probability slice into each channel of summed_probabilities given by branch.
                # This duplicates probabilities for easy passing to standard loss functions such as nll_loss.
                for channel in branch:
                    summed_probabilities[:, channel : (channel + 1), :, :] = summed_tree_branch_slice
            loss = self.get_loss(loss, summed_probabilities, target)
        return loss


class WeightedHierarchicalLoss(HierarchicalLoss):
    def get_loss(self, running_loss, summed_probabilities, target):
        return super().get_loss(running_loss, summed_probabilities, target) * (1 / (self.n_level + 1))


class ConditionalHierarchicalLoss(HierarchicalLoss):
    def forward(self, logits: Tensor, target: Tensor) -> Tensor:
        eps = 1e-4
        last_summed_probabilities = torch.Tensor()
        previous_branch_for_channel = {}
        probabilities = F.softmax(logits, dim=1)
        loss = 0
        precomputed_hierarchy_list = self.root.loss_layers
        for n_level, level_loss_list in enumerate(precomputed_hierarchy_list):
            self.n_level = n_level
            probabilities_tosum = probabilities.clone()
            summed_probabilities = probabilities_tosum
            for branch in level_loss_list:
                branch_probs = torch.FloatTensor()
                branch_probs = branch_probs.to(logits.device)
                for channel in branch:
                    branch_probs = torch.cat((branch_probs, probabilities_tosum[:, channel].unsqueeze(1)), 1)

                summed_tree_branch_slice = branch_probs.sum(1, keepdim=True)

                for channel in branch:
                    if channel not in previous_branch_for_channel:
                        # First iteration
                        previous_branch_for_channel[channel] = branch

                    if previous_branch_for_channel[channel] != branch:
                        summed_probabilities[:, channel : (channel + 1), :, :] = summed_tree_branch_slice * (
                            last_summed_probabilities[:, channel : (channel + 1), :, :] + eps
                        )
                        previous_branch_for_channel[channel] = branch
                    else:
                        if len(last_summed_probabilities) > 0:
                            summed_probabilities[:, channel : (channel + 1), :, :] = last_summed_probabilities[
                                :, channel : (channel + 1), :, :
                            ]
                        else:
                            summed_probabilities[:, channel : (channel + 1), :, :] = summed_tree_branch_slice

            loss = self.get_loss(loss, summed_probabilities, target, only_last=True)
            last_summed_probabilities = summed_probabilities
        return loss


class HierarchicalDiceCELoss(HierarchicalLoss):
    def __init__(self, hierarchical_kwargs={}):
        super(HierarchicalDiceCELoss, self).__init__(hierarchical_kwargs)
        self.loss_fn_kwargs = {"soft_dice_kwargs": {"apply_softmax": False}}
        self.loss = DiceCE(**self.loss_fn_kwargs)


class WeightedHierarchicalDiceCELoss(HierarchicalDiceCELoss, WeightedHierarchicalLoss):
    def __init__(self, hierarchical_kwargs={}):
        super().__init__(hierarchical_kwargs)


class ConditionalHierarchicalDiceCELoss(HierarchicalDiceCELoss, ConditionalHierarchicalLoss):
    def __init__(self, hierarchical_kwargs={}):
        super().__init__(hierarchical_kwargs)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from yucca.network_architectures.blocks_and_layers.conv_layers import (
    ConvDropoutNormNonlin,
    ConvDropoutNorm,
)
from yucca.network_architectures.blocks_and_layers.norm import LayerNorm3d
from timm.layers import DropPath
from functools import partial
from typing import Callable, Optional


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition" https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = nn.Conv2d(inplanes, width, kernel_size=1, stride=1, bias=False)
        self.bn1 = norm_layer(width)
        self.conv2 = nn.Conv2d(
            width,
            width,
            kernel_size=3,
            stride=stride,
            groups=groups,
            dilation=dilation,
            padding=dilation,
            bias=False,
        )
        self.bn2 = norm_layer(width)
        self.conv3 = nn.Conv2d(width, planes * self.expansion, kernel_size=1, stride=1, bias=False)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class MedNeXtBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        exp_r: int = 4,
        kernel_size: int = 7,
        do_res: int = True,
        norm_type: str = "group",
        n_groups: int or None = None,
        dim="3d",
        grn=False,
    ):
        super().__init__()

        self.do_res = do_res

        assert dim in ["2d", "3d"]
        self.dim = dim
        if self.dim == "2d":
            conv = nn.Conv2d
        elif self.dim == "3d":
            conv = nn.Conv3d

        # First convolution layer with DepthWise Convolutions
        self.conv1 = conv(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=1,
            padding=kernel_size // 2,
            groups=in_channels if n_groups is None else n_groups,
        )

        # Normalization Layer. GroupNorm is used by default.
        if norm_type == "group":
            self.norm = nn.GroupNorm(num_groups=in_channels, num_channels=in_channels)
        elif norm_type == "layer":
            self.norm = LayerNorm(normalized_shape=in_channels, data_format="channels_first")

        # Second convolution (Expansion) layer with Conv3D 1x1x1
        self.conv2 = conv(
            in_channels=in_channels,
            out_channels=exp_r * in_channels,
            kernel_size=1,
            stride=1,
            padding=0,
        )

        # GeLU activations
        self.act = nn.GELU()

        # Third convolution (Compression) layer with Conv3D 1x1x1
        self.conv3 = conv(
            in_channels=exp_r * in_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
            padding=0,
        )

        self.grn = grn
        if grn:
            if dim == "3d":
                self.grn_beta = nn.Parameter(torch.zeros(1, exp_r * in_channels, 1, 1, 1), requires_grad=True)
                self.grn_gamma = nn.Parameter(torch.zeros(1, exp_r * in_channels, 1, 1, 1), requires_grad=True)
            elif dim == "2d":
                self.grn_beta = nn.Parameter(torch.zeros(1, exp_r * in_channels, 1, 1), requires_grad=True)
                self.grn_gamma = nn.Parameter(torch.zeros(1, exp_r * in_channels, 1, 1), requires_grad=True)

    def forward(self, x, _dummy_tensor=None):
        x1 = x
        x1 = self.conv1(x1)
        x1 = self.act(self.conv2(self.norm(x1)))
        if self.grn:
            # gamma, beta: learnable affine transform parameters
            # X: input of shape (N,C,H,W,D)
            if self.dim == "3d":
                gx = torch.norm(x1, p=2, dim=(-3, -2, -1), keepdim=True)
            elif self.dim == "2d":
                gx = torch.norm(x1, p=2, dim=(-2, -1), keepdim=True)
            nx = gx / (gx.mean(dim=1, keepdim=True) + 1e-6)
            x1 = self.grn_gamma * (x1 * nx) + self.grn_beta + x1
        x1 = self.conv3(x1)
        if self.do_res:
            x1 = x + x1
        return x1


class MedNeXtDownBlock(MedNeXtBlock):
    def __init__(
        self,
        in_channels,
        out_channels,
        exp_r=4,
        kernel_size=7,
        do_res=False,
        norm_type="group",
        dim="3d",
        grn=False,
    ):
        super().__init__(
            in_channels,
            out_channels,
            exp_r,
            kernel_size,
            do_res=False,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        if dim == "2d":
            conv = nn.Conv2d
        elif dim == "3d":
            conv = nn.Conv3d
        self.resample_do_res = do_res
        if do_res:
            self.res_conv = conv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=2,
            )

        self.conv1 = conv(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=2,
            padding=kernel_size // 2,
            groups=in_channels,
        )

    def forward(self, x, _dummy_tensor=None):
        x1 = super().forward(x)

        if self.resample_do_res:
            res = self.res_conv(x)
            x1 = x1 + res

        return x1


class MedNeXtUpBlock(MedNeXtBlock):
    def __init__(
        self,
        in_channels,
        out_channels,
        exp_r=4,
        kernel_size=7,
        do_res=False,
        norm_type="group",
        dim="3d",
        grn=False,
    ):
        super().__init__(
            in_channels,
            out_channels,
            exp_r,
            kernel_size,
            do_res=False,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.resample_do_res = do_res

        self.dim = dim
        if dim == "2d":
            conv = nn.ConvTranspose2d
        elif dim == "3d":
            conv = nn.ConvTranspose3d
        if do_res:
            self.res_conv = conv(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=2,
            )

        self.conv1 = conv(
            in_channels=in_channels,
            out_channels=in_channels,
            kernel_size=kernel_size,
            stride=2,
            padding=kernel_size // 2,
            groups=in_channels,
        )

    def forward(self, x, _dummy_tensor=None):
        x1 = super().forward(x)
        # Asymmetry but necessary to match shape

        if self.dim == "2d":
            x1 = torch.nn.functional.pad(x1, (1, 0, 1, 0))
        elif self.dim == "3d":
            x1 = torch.nn.functional.pad(x1, (1, 0, 1, 0, 1, 0))

        if self.resample_do_res:
            res = self.res_conv(x)
            if self.dim == "2d":
                res = torch.nn.functional.pad(res, (1, 0, 1, 0))
            elif self.dim == "3d":
                res = torch.nn.functional.pad(res, (1, 0, 1, 0, 1, 0))
            x1 = x1 + res

        return x1


class OutBlock(nn.Module):
    def __init__(self, in_channels, n_classes, dim):
        super().__init__()

        if dim == "2d":
            conv = nn.ConvTranspose2d
        elif dim == "3d":
            conv = nn.ConvTranspose3d
        self.conv_out = conv(in_channels, n_classes, kernel_size=1)

    def forward(self, x, _dummy_tensor=None):
        return self.conv_out(x)


class LayerNorm(nn.Module):
    """LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """

    def __init__(self, normalized_shape, eps=1e-5, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))  # beta
        self.bias = nn.Parameter(torch.zeros(normalized_shape))  # gamma
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x, _dummy_tensor=False):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None, None] * x + self.bias[:, None, None, None]
            return x


class Multiresblock(nn.Module):
    """
    MultiRes Block

    Arguments:
    num_in_channels {int} -- Number of channels coming into mutlires block
    num_filters {int} -- Number of filters in a corrsponding UNet stage
    alpha {float} -- alpha hyperparameter (default: 1.67)
    """

    def __init__(
        self,
        num_in_channels,
        num_filters,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        basic_block=ConvDropoutNormNonlin,
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        self.alpha = 1.67
        self.W = num_filters * self.alpha

        filt_cnt_3x3 = int(self.W * 0.167)
        filt_cnt_5x5 = int(self.W * 0.333)
        filt_cnt_7x7 = int(self.W * 0.5)
        num_out_filters = filt_cnt_3x3 + filt_cnt_5x5 + filt_cnt_7x7

        shortcut_kwargs = {"kernel_size": 1}

        self.shortcut = ConvDropoutNorm(
            num_in_channels,
            num_out_filters,
            self.conv_op,
            shortcut_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.conv_3x3 = basic_block(
            num_in_channels,
            filt_cnt_3x3,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.conv_5x5 = basic_block(
            filt_cnt_3x3,
            filt_cnt_5x5,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.conv_7x7 = basic_block(
            filt_cnt_5x5,
            filt_cnt_7x7,
            self.conv_op,
            self.conv_kwargs,
            self.norm_op,
            self.norm_op_kwargs,
            self.dropout_op,
            self.dropout_op_kwargs,
            self.nonlin,
            self.nonlin_kwargs,
        )

        self.norm1 = self.norm_op(num_out_filters, **self.norm_op_kwargs)
        self.norm2 = self.norm_op(num_out_filters, **self.norm_op_kwargs)
        self.activation = self.nonlin(**self.nonlin_kwargs)

    def forward(self, x):
        shrtct = self.shortcut(x)

        a = self.conv_3x3(x)
        b = self.conv_5x5(a)
        c = self.conv_7x7(b)

        x = torch.cat([a, b, c], axis=1)
        x = self.norm1(x)

        x = x + shrtct
        x = self.norm2(x)
        x = self.activation(x)

        return x


class Respath(torch.nn.Module):
    """
    ResPath

    Arguments:
    num_in_filters {int} -- Number of filters going in the respath
    num_out_filters {int} -- Number of filters going out the respath
    respath_length {int} -- length of ResPath

    """

    def __init__(
        self,
        num_in_filters,
        num_out_filters,
        respath_length,
        conv_op=nn.Conv2d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=nn.BatchNorm2d,
        norm_op_kwargs={"eps": 1e-5, "affine": True, "momentum": 0.1},
        dropout_op=nn.Dropout2d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.LeakyReLU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        basic_block=ConvDropoutNormNonlin,
    ):
        super().__init__()

        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.conv_kwargs = conv_kwargs
        self.conv_op = conv_op
        self.norm_op = norm_op

        shortcut_kwargs = {"kernel_size": 1}

        self.respath_length = respath_length
        self.shortcuts = torch.nn.ModuleList([])
        self.convs = torch.nn.ModuleList([])
        self.bns = torch.nn.ModuleList([])
        self.activation = self.nonlin(**self.nonlin_kwargs)

        for i in range(self.respath_length):
            if i == 0:
                self.shortcuts.append(
                    ConvDropoutNorm(
                        num_in_filters,
                        num_out_filters,
                        self.conv_op,
                        shortcut_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    )
                )
                self.convs.append(
                    basic_block(
                        num_in_filters,
                        num_out_filters,
                        self.conv_op,
                        self.conv_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    )
                )

            else:
                self.shortcuts.append(
                    ConvDropoutNorm(
                        num_out_filters,
                        num_out_filters,
                        self.conv_op,
                        shortcut_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    )
                )

                self.convs.append(
                    basic_block(
                        num_out_filters,
                        num_out_filters,
                        self.conv_op,
                        self.conv_kwargs,
                        self.norm_op,
                        self.norm_op_kwargs,
                        self.dropout_op,
                        self.dropout_op_kwargs,
                        self.nonlin,
                        self.nonlin_kwargs,
                    )
                )

            self.bns.append(self.norm_op(num_out_filters, **self.norm_op_kwargs))

    def forward(self, x):
        for i in range(self.respath_length):
            shortcut = self.shortcuts[i](x)

            x = self.convs[i](x)
            x = self.bns[i](x)
            x = self.activation(x)

            x = x + shortcut
            x = self.bns[i](x)
            x = self.activation(x)
        return x


class ux_block(nn.Module):
    r"""ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch

    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(
        self,
        dim,
        drop_path=0.0,
        layer_scale_init_value=1e-6,
        conv_op=nn.Conv3d,
        norm_op=LayerNorm3d,
        norm_op_kwargs={"eps": 1e-6},
        nonlin=nn.GELU,
    ):
        super().__init__()
        if conv_op == nn.Conv2d:
            self.permute_fw = [0, 2, 3, 1]
            self.permute_bw = [0, 3, 1, 2]
        else:
            self.permute_fw = [0, 2, 3, 4, 1]
            self.permute_bw = [0, 4, 1, 2, 3]

        self.dwconv = conv_op(dim, dim, kernel_size=7, padding=3, groups=dim)
        self.norm = norm_op(dim, **norm_op_kwargs)
        self.pwconv1 = conv_op(dim, 4 * dim, kernel_size=1, groups=dim)
        self.nonlin = nonlin()
        self.pwconv2 = conv_op(4 * dim, dim, kernel_size=1, groups=dim)
        self.gamma = (
            nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
            if layer_scale_init_value > 0
            else None
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x):
        input = x
        x = self.dwconv(x)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.nonlin(x)
        x = self.pwconv2(x)

        if self.gamma is not None:
            x = x.permute(self.permute_fw)
            x = self.gamma * x
            x = x.permute(self.permute_bw)

        x = input + self.drop_path(x)
        return x


class UXNet_encoder(nn.Module):
    """
    Args:
        in_channels (int): Number of input image channels. Default: 3
        num_classes (int): Number of classes for classification head. Default: 1000
        depths (tuple(int)): Number of blocks at each stage. Default: [3, 3, 9, 3]
        dims (int): Feature dimension at each stage. Default: [96, 192, 384, 768]
        drop_path_rate (float): Stochastic depth rate. Default: 0.
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
        head_init_scale (float): Init scaling value for classifier weights and biases. Default: 1.
    """

    def __init__(
        self,
        in_channels=1,
        depths=[2, 2, 2, 2],
        dims=[48, 96, 192, 384],
        drop_path_rate=0.0,
        layer_scale_init_value=1e-6,
        out_indices=[0, 1, 2, 3],
        conv_op=nn.Conv3d,
        conv_kwargs={
            "kernel_size": 3,
            "stride": 1,
            "padding": 1,
            "dilation": 1,
            "bias": True,
        },
        norm_op=LayerNorm3d,
        norm_op_kwargs={"eps": 1e-6},
        dropout_op=nn.Dropout3d,
        dropout_op_kwargs={"p": 0.5, "inplace": True},
        nonlin=nn.GELU,
        nonlin_kwargs={"negative_slope": 1e-2, "inplace": True},
        basic_block=ux_block,
    ):
        super().__init__()

        # Model parameters
        self.conv_op = conv_op
        self.conv_kwargs = conv_kwargs
        self.norm_op_kwargs = norm_op_kwargs
        self.norm_op = norm_op
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.nonlin_kwargs = nonlin_kwargs
        self.nonlin = nonlin
        self.basic_block = basic_block

        self.downsample_layers = nn.ModuleList()  # stem and 3 intermediate downsampling conv layers
        stem = nn.Sequential(
            self.conv_op(in_channels, dims[0], kernel_size=7, stride=2, padding=3),
            self.norm_op(dims[0], **self.norm_op_kwargs),
        )
        self.downsample_layers.append(stem)
        for i in range(3):
            downsample_layer = nn.Sequential(
                self.norm_op(dims[i], **self.norm_op_kwargs),
                self.conv_op(dims[i], dims[i + 1], kernel_size=2, stride=2),
            )
            self.downsample_layers.append(downsample_layer)

        self.stages = nn.ModuleList()
        dp_rates = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        cur = 0
        for i in range(4):
            stage = nn.Sequential(
                *[
                    self.basic_block(
                        dim=dims[i],
                        conv_op=self.conv_op,
                        norm_op=self.norm_op,
                        drop_path=dp_rates[cur + j],
                        layer_scale_init_value=layer_scale_init_value,
                    )
                    for j in range(depths[i])
                ]
            )

            self.stages.append(stage)
            cur += depths[i]

        self.out_indices = out_indices

        norm_layer = partial(self.norm_op, eps=self.norm_op_kwargs["eps"])
        for i_layer in range(4):
            layer = norm_layer(dims[i_layer])
            layer_name = f"norm{i_layer}"
            self.add_module(layer_name, layer)

        # self.apply(self._init_weights)

    def forward_features(self, x):
        outs = []
        for i in range(4):
            x = self.downsample_layers[i](x)
            x = self.stages[i](x)
            if i in self.out_indices:
                norm_layer = getattr(self, f"norm{i}")
                x_out = norm_layer(x)
                outs.append(x_out)

        return tuple(outs)

    def forward(self, x):
        x = self.forward_features(x)
        return x



import torch
import torch.nn as nn


class LayerNorm3d(nn.Module):
    def __init__(self, normalized_shape, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        # print(self.weight.size())
        x = self.weight[:, None, None, None] * x + self.bias[:, None, None, None]

        return x


class LayerNorm2d(nn.Module):
    def __init__(self, normalized_shape, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        u = x.mean(1, keepdim=True)
        s = (x - u).pow(2).mean(1, keepdim=True)
        x = (x - u) / torch.sqrt(s + self.eps)
        # print(self.weight.size())
        x = self.weight[:, None, None] * x + self.bias[:, None, None]

        return x



import nibabel as nib
import numpy as np
from sklearn.model_selection import train_test_split
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
from yucca.utils.nib_utils import get_nib_orientation, get_nib_spacing


def convert(path: str, subdir: str = "decathlon", subsubdir: str = "Task03_Liver"):
    # INPUT DATA
    path = f"{path}/{subdir}/{subsubdir}"

    file_suffix = ".nii.gz"

    # Train/Test Splits
    images_dir = join(path, "imagesTr")
    labels_dir = join(path, "labelsTr")
    training_samples, test_samples = train_test_split(subfiles(labels_dir, suffix=file_suffix, join=False), random_state=333)

    ###OUTPUT DATA
    # Target names
    task_name = "Task023_Decathlon_Liver"
    task_prefix = "Decathlon_Liver"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    images_dir_tr = images_dir_ts = images_dir
    labels_dir_tr = labels_dir_ts = labels_dir

    skipped = []

    for sTr in training_samples:
        image = nib.load(join(images_dir_tr, sTr))
        label = nib.load(join(labels_dir_tr, sTr))
        sTr = sTr[: -len(file_suffix)]

        # Orient to RAS and register image-label, using the image as reference.
        orig_ornt = get_nib_orientation(image)
        label_ornt = get_nib_orientation(label)

        if orig_ornt == label_ornt:

            print("same orientation")

            orig_spacing = get_nib_spacing(image)
            label_spacing = get_nib_spacing(label)

            if orig_spacing.all() == label_spacing.all():

                print("same spacing")

                orig_shape = np.shape(image)
                label_shape = np.shape(label)

                if orig_shape == label_shape:

                    nib.save(image, filename=f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz")
                    nib.save(label, filename=f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz")
                    print("same shape, file saved")

                else:
                    print("shape not matching, file not saved")
                    skipped.append(join(images_dir_tr, sTr))

            else:

                print("Spacing not matching")
                skipped.append(join(images_dir_tr, sTr))

        else:
            print("Orientation not matching")
            skipped.append(join(images_dir_tr, sTr))

    for sTs in test_samples:
        image = nib.load(join(images_dir_ts, sTs))
        label = nib.load(join(labels_dir_ts, sTs))
        sTs = sTs[: -len(file_suffix)]

        # Orient to RAS and register image-label, using the image as reference.
        orig_ornt = get_nib_orientation(image)
        label_ornt = get_nib_orientation(label)

        if orig_ornt == label_ornt:

            print("same orientation")

            orig_spacing = get_nib_spacing(image)
            label_spacing = get_nib_spacing(label)

            if orig_spacing.all() == label_spacing.all():

                print("same spacing")

                orig_shape = np.shape(image)
                label_shape = np.shape(label)

                if orig_shape == label_shape:

                    nib.save(image, filename=f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz")
                    nib.save(label, filename=f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz")
                    print("same shape, file saved")

                else:
                    print("shape not matching, file not saved")
                    skipped.append(join(images_dir_ts, sTs))

            else:

                print("Spacing not matching")
                skipped.append(join(images_dir_ts, sTs))

        else:
            print("Orientation not matching")
            skipped.append(join(images_dir_ts, sTs))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("CT",),
        labels={0: "Background", 1: "Liver", 2: "Cancer"},
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Liver",
        dataset_reference="King's College London",
    )



from torchvision import transforms
from yucca.image_processing.matrix_ops import get_max_rotated_size
from yucca.image_processing.transforms.formatting import (
    AddBatchDimension,
    RemoveBatchDimension,
)
from yucca.image_processing.transforms.BiasField import BiasField
from yucca.image_processing.transforms.Blur import Blur
from yucca.image_processing.transforms.CopyImageToSeg import CopyImageToSeg
from yucca.image_processing.transforms.Gamma import Gamma
from yucca.image_processing.transforms.Ghosting import MotionGhosting
from yucca.image_processing.transforms.Masking import Masking
from yucca.image_processing.transforms.Mirror import Mirror
from yucca.image_processing.transforms.Noise import (
    AdditiveNoise,
    MultiplicativeNoise,
)
from yucca.image_processing.transforms.Ringing import GibbsRinging
from yucca.image_processing.transforms.sampling import DownsampleSegForDS
from yucca.image_processing.transforms.SimulateLowres import SimulateLowres
from yucca.image_processing.transforms.Spatial import Spatial


class YuccaAugmentationComposer:
    def __init__(
        self,
        patch_size: list | tuple,
        deep_supervision: bool = False,
        is_2D: bool = False,
        parameter_dict: dict = {},
        task_type_preset: str = "segmentation",
    ):
        self._pre_aug_patch_size = None
        self.deep_supervision = deep_supervision
        self.setup_default_params(is_2D, patch_size)
        self.apply_task_type_specific_preset(task_type_preset)
        self.overwrite_params(parameter_dict)
        self.train_transforms = self.compose_train_transforms()
        self.val_transforms = self.compose_val_transforms()

    def setup_default_params(self, is_2D, patch_size):
        print("Composing Transforms")
        # Define whether we crop before or after applying augmentations
        # Define if the cropping performed during the spatial transform is random or always centered
        #   We have ALREADY selected a random crop when we prepared the case in the dataloader
        #   so unless you know what you're doing this should be disabled to avoid border artifacts
        self.random_crop = False
        self.mask_image_for_reconstruction = False
        self.patch_size = patch_size
        self.cval = "min"  # can be an int, float or a str in ['min', 'max']

        # label/segmentation transforms
        self.skip_label = False
        self.label_dtype = int
        self.copy_image_to_label = False

        # Default probabilities (all ON by default)
        # For augmentations with AUG_P_PER_SAMPLE, AUG_P_PER_CHANNEL and AUG_P_PER_AXIS they are applied in the following order:
        #
        # for sample in dataset: # iterate over the batch
        #  if p_per_sample > np.random.uniform():
        #    for channel in sample: # iterate over the channel/modalities
        #      if p_per_channel > np.random.uniform():
        #        for axis in channel: # iterate over the h,w,d dimensions
        #          if p_per_axis > np.random.uniform():
        #            sample[channel, axis] = apply_aug(sample[channel, axis])
        # Therefore, to enable any augmentation the AUG_P_PER_SAMPLE must be > 0.

        self.additive_noise_p_per_sample = 0.2
        self.biasfield_p_per_sample = 0.33
        self.blurring_p_per_sample = 0.2
        self.blurring_p_per_channel = 0.5
        self.elastic_deform_p_per_sample = 0.33
        self.gamma_p_per_sample = 0.2
        self.gamma_p_invert_image = 0.05
        self.gibbs_ringing_p_per_sample = 0.2
        self.mirror_p_per_sample = 0.0
        self.mirror_p_per_axis = 0.33
        self.motion_ghosting_p_per_sample = 0.2
        self.multiplicative_noise_p_per_sample = 0.2
        self.rotation_p_per_sample = 0.2
        self.rotation_p_per_axis = 0.66
        self.scale_p_per_sample = 0.2
        self.simulate_lowres_p_per_sample = 0.2
        self.simulate_lowres_p_per_channel = 0.5
        self.simulate_lowres_p_per_axis = 0.33

        # default augmentation values
        self.additive_noise_mean = (0.0, 0.0)
        self.additive_noise_sigma = (1e-3, 1e-4)

        self.blurring_sigma = (0.0, 1.0)

        self.elastic_deform_alpha = (200, 600)
        self.elastic_deform_sigma = (20, 30)

        self.gamma_range = (0.5, 2.0)

        self.gibbs_ringing_cutfreq = (96, 129)
        self.gibbs_ringing_axes = (0, 2) if is_2D else (0, 3)

        self.mask_ratio = 0.5

        self.mirror_axes = (0, 1) if is_2D else (0, 1, 2)

        self.motion_ghosting_alpha = (0.85, 0.95)
        self.motion_ghosting_numreps = (2, 11)
        self.motion_ghosting_axes = (0, 2) if is_2D else (0, 3)

        self.multiplicative_noise_mean = (0, 0)
        self.multiplicative_noise_sigma = (1e-3, 1e-4)

        self.rotation_x = (-30.0, 30.0)
        self.rotation_y = (-0.0, 0.0) if is_2D else (-30.0, 30.0)
        self.rotation_z = (-0.0, 0.0) if is_2D else (-30.0, 30.0)

        self.scale_factor = (0.9, 1.1)

        self.simulate_lowres_zoom_range = (0.5, 1.0)

    @property
    def pre_aug_patch_size(self):
        # First check if any spatial transforms are included
        if self.elastic_deform_p_per_sample > 0 or self.rotation_p_per_sample > 0 or self.scale_p_per_sample > 0:
            self._pre_aug_patch_size = get_max_rotated_size(self.patch_size)
        return self._pre_aug_patch_size

    def apply_task_type_specific_preset(self, preset):
        if preset == "segmentation":
            # we do nothing and use the default parameters
            pass

        elif preset == "classification":
            self.skip_label = True

        elif preset == "self-supervised":
            self.skip_label = True
            self.copy_image_to_label = True
            self.mask_image_for_reconstruction = True

        else:
            raise ValueError(f"{preset} is not a valid `task_type_preset`.")

    def overwrite_params(self, parameter_dict):
        for key, value in parameter_dict.items():
            setattr(self, key, value)

    def compose_train_transforms(self):
        tr_transforms = transforms.Compose(
            [
                AddBatchDimension(),
                Spatial(
                    patch_size=self.patch_size,
                    crop=True,
                    random_crop=self.random_crop,
                    cval=self.cval,
                    p_deform_per_sample=self.elastic_deform_p_per_sample,
                    deform_sigma=self.elastic_deform_sigma,
                    deform_alpha=self.elastic_deform_alpha,
                    p_rot_per_sample=self.rotation_p_per_sample,
                    p_rot_per_axis=self.rotation_p_per_axis,
                    x_rot_in_degrees=self.rotation_x,
                    y_rot_in_degrees=self.rotation_y,
                    z_rot_in_degrees=self.rotation_z,
                    p_scale_per_sample=self.scale_p_per_sample,
                    scale_factor=self.scale_factor,
                    skip_label=self.skip_label,
                ),
                AdditiveNoise(
                    p_per_sample=self.additive_noise_p_per_sample,
                    mean=self.additive_noise_mean,
                    sigma=self.additive_noise_sigma,
                ),
                Blur(
                    p_per_sample=self.blurring_p_per_sample,
                    p_per_channel=self.blurring_p_per_channel,
                    sigma=self.blurring_sigma,
                ),
                MultiplicativeNoise(
                    p_per_sample=self.multiplicative_noise_p_per_sample,
                    mean=self.multiplicative_noise_mean,
                    sigma=self.multiplicative_noise_sigma,
                ),
                MotionGhosting(
                    p_per_sample=self.motion_ghosting_p_per_sample,
                    alpha=self.motion_ghosting_alpha,
                    numReps=self.motion_ghosting_numreps,
                    axes=self.motion_ghosting_axes,
                ),
                GibbsRinging(
                    p_per_sample=self.gibbs_ringing_p_per_sample,
                    cutFreq=self.gibbs_ringing_cutfreq,
                    axes=self.gibbs_ringing_axes,
                ),
                SimulateLowres(
                    p_per_sample=self.simulate_lowres_p_per_sample,
                    p_per_channel=self.simulate_lowres_p_per_channel,
                    p_per_axis=self.simulate_lowres_p_per_axis,
                    zoom_range=self.simulate_lowres_zoom_range,
                ),
                BiasField(p_per_sample=self.biasfield_p_per_sample),
                Gamma(
                    p_per_sample=self.gamma_p_per_sample,
                    p_invert_image=self.gamma_p_invert_image,
                    gamma_range=self.gamma_range,
                ),
                Mirror(
                    p_per_sample=self.mirror_p_per_sample,
                    axes=self.mirror_axes,
                    p_mirror_per_axis=self.mirror_p_per_axis,
                    skip_label=self.skip_label,
                ),
                DownsampleSegForDS(deep_supervision=self.deep_supervision),
                CopyImageToSeg(copy=self.copy_image_to_label),
                Masking(mask=self.mask_image_for_reconstruction, pixel_value=self.cval, ratio=self.mask_ratio),
                RemoveBatchDimension(),
            ]
        )
        return tr_transforms

    def compose_val_transforms(self):
        val_transforms = transforms.Compose(
            [
                AddBatchDimension(),
                CopyImageToSeg(copy=self.copy_image_to_label),
                Masking(mask=self.mask_image_for_reconstruction, pixel_value=self.cval, ratio=self.mask_ratio),
                RemoveBatchDimension(),
            ]
        )
        return val_transforms

    def lm_hparams(self):
        hparams = {
            "augmentation_parameters": {
                "deep_supervision": self.deep_supervision,
                "pre_aug_patch_size": self.pre_aug_patch_size,
                "random_crop": self.random_crop,
                "cval": self.cval,
                "mask_image_for_reconstruction": self.mask_image_for_reconstruction,
                "patch_size": self.patch_size,
                "skip_label": self.skip_label,
                "label_dtype": self.label_dtype,
                "copy_image_to_label": self.copy_image_to_label,
                "additive_noise_p_per_sample": self.additive_noise_p_per_sample,
                "additive_noise_mean": self.additive_noise_mean,
                "additive_noise_sigma": self.additive_noise_sigma,
                "biasfield_p_per_sample": self.biasfield_p_per_sample,
                "blurring_p_per_sample": self.blurring_p_per_sample,
                "blurring_sigma": self.blurring_sigma,
                "blurring_p_per_channel": self.blurring_p_per_channel,
                "elastic_deform_p_per_sample": self.elastic_deform_p_per_sample,
                "elastic_deform_alpha": self.elastic_deform_alpha,
                "elastic_deform_sigma": self.elastic_deform_sigma,
                "gamma_p_per_sample": self.gamma_p_per_sample,
                "gamma_p_invert_image": self.gamma_p_invert_image,
                "gamma_range": self.gamma_range,
                "gibbs_ringing_p_per_sample": self.gibbs_ringing_p_per_sample,
                "gibbs_ringing_cutfreq": self.gibbs_ringing_cutfreq,
                "gibbs_ringing_axes": self.gibbs_ringing_axes,
                "mask_ratio": self.mask_ratio,
                "mirror_p_per_sample": self.mirror_p_per_sample,
                "mirror_p_per_axis": self.mirror_p_per_axis,
                "mirror_axes": self.mirror_axes,
                "motion_ghosting_p_per_sample": self.motion_ghosting_p_per_sample,
                "motion_ghosting_alpha": self.motion_ghosting_alpha,
                "motion_ghosting_numreps": self.motion_ghosting_numreps,
                "motion_ghosting_axes": self.motion_ghosting_axes,
                "multiplicative_noise_p_per_sample": self.multiplicative_noise_p_per_sample,
                "multiplicative_noise_mean": self.multiplicative_noise_mean,
                "multiplicative_noise_sigma": self.multiplicative_noise_sigma,
                "rotation_p_per_sample": self.rotation_p_per_sample,
                "rotation_p_per_axis": self.rotation_p_per_axis,
                "rotation_x": self.rotation_x,
                "rotation_y": self.rotation_y,
                "rotation_z": self.rotation_z,
                "scale_p_per_sample": self.scale_p_per_sample,
                "scale_factor": self.scale_factor,
                "simulate_lowres_p_per_sample": self.simulate_lowres_p_per_sample,
                "simulate_lowres_p_per_channel": self.simulate_lowres_p_per_channel,
                "simulate_lowres_p_per_axis": self.simulate_lowres_p_per_axis,
                "simulate_lowres_zoom_range": self.simulate_lowres_zoom_range,
            }
        }
        return hparams


if __name__ == "__main__":
    from yucca.training.augmentation.augmentation_presets import basic

    x = YuccaAugmentationComposer(patch_size=(32, 32), parameter_dict=basic)
    print("ALL AUGMENTATION PARAMETERS: ", x.lm_hparams())
    print("")
    print("BASIC PARAMETERS: ", basic)



import os
from yucca.task_conversion.utils import combine_images_from_tasks, generate_dataset_json
from yucca.paths import yucca_raw_data


def convert(_path: str, _subdir: str = None):
    # Define the name of the new task
    task_name = "Task245_BRAINS-45K"

    # Define the expected labels (leave empty for no labels)
    expected_labels = {}

    # Define the tasks to combine, such as ["Task001_OASIS", "Task002_LPBA40"]
    # The individual task_conversion scripts must be run prior to executing this, as the script will look for the
    # data in the yucca_raw_data folder.
    tasks_to_combine = [
        "Task201_PPMI",
        "Task202_ISLES22",
        "Task203_OASIS3",
        "Task204_OASIS4",
        "Task206_BrainTumour",
        "Task207_ADNI",
        "Task208_WMH",
        "Task209_BraTS21",
        "Task210_MSSEG1",
    ]

    target_base = os.path.join(yucca_raw_data, task_name)
    os.makedirs(target_base, exist_ok=True)

    combine_images_from_tasks(tasks=tasks_to_combine, target_base=target_base, run_type="unsupervised")

    generate_dataset_json(
        os.path.join(target_base, "dataset.json"),
        imagesTr_dir=os.path.join(target_base, "imagesTr"),
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=expected_labels,
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Combination of multiple datasets",
        dataset_reference="",
    )



import lightning as L
import torch
import yucca
import wandb
import copy
import logging
import numpy as np
import os
import matplotlib.pyplot as plt
from batchgenerators.utilities.file_and_folder_operations import join
from torchmetrics import MetricCollection
from torchmetrics.classification import Dice, Accuracy, AUROC
from torchmetrics.regression import MeanAbsoluteError
from yucca.training.loss_and_optim.loss_functions.deep_supervision import DeepSupervisionLoss
from yucca.utils.files_and_folders import recursive_find_python_class
from yucca.utils.kwargs import filter_kwargs
from yucca.evaluation.training_metrics import F1


class YuccaLightningModule(L.LightningModule):
    """
    The YuccaLightningModule class is an implementation of the PyTorch Lightning module designed for the Yucca project.
    It extends the LightningModule class and encapsulates the neural network model, loss functions, and optimization logic.
    This class is responsible for handling training, validation, and inference steps within the Yucca machine learning pipeline.


    """

    def __init__(
        self,
        config: dict,
        deep_supervision: bool = False,
        learning_rate: float = 1e-3,
        loss_fn: str = None,
        lr_scheduler: torch.optim.lr_scheduler._LRScheduler = torch.optim.lr_scheduler.CosineAnnealingLR,
        momentum: float = 0.9,
        optimizer: torch.optim.Optimizer = torch.optim.SGD,
        sliding_window_overlap: float = 0.5,
        step_logging: bool = False,
        test_time_augmentation: bool = False,
        progress_bar: bool = False,
    ):
        super().__init__()
        # Extract parameters from the configurator
        self.num_classes = config["num_classes"]
        self.num_modalities = config["num_modalities"]
        self.version_dir = config["version_dir"]
        self.plans = config["plans"]
        self.plans_path = config["plans_path"]
        self.model_name = config["model_name"]
        self.model_dimensions = config["model_dimensions"]
        self.patch_size = config["patch_size"]
        self.task_type = config["task_type"]
        self.sliding_window_prediction = config["patch_based_training"]

        # Loss, optimizer and scheduler parameters
        self.deep_supervision = deep_supervision
        self.lr = learning_rate
        self.loss_fn = loss_fn

        self.momentum = momentum
        self.optim = optimizer
        self.lr_scheduler = lr_scheduler

        # Evaluation and logging
        if step_logging is True:  # Blame PyTorch lightning for this war crime
            self.step_logging = None
            self.epoch_logging = None
        else:
            self.step_logging = False
            self.epoch_logging = True

        self.progress_bar = progress_bar

        if self.task_type == "classification":
            tmetrics_task = "multiclass" if self.num_classes > 2 else "binary"
            # can we get per-class?
            self.train_metrics = MetricCollection(
                {
                    "train_acc": Accuracy(task=tmetrics_task, num_classes=self.num_classes),
                    "train_roc_auc": AUROC(task=tmetrics_task, num_classes=self.num_classes),
                }
            )
            self.val_metrics = MetricCollection(
                {
                    "val_acc": Accuracy(task=tmetrics_task, num_classes=self.num_classes),
                    "val_roc_auc": AUROC(task=tmetrics_task, num_classes=self.num_classes),
                }
            )

        if self.task_type == "segmentation":
            self.train_metrics = MetricCollection(
                {
                    "train/dice": Dice(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None),
                    "train/F1": F1(
                        num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None, average=None
                    ),
                },
            )

            self.val_metrics = MetricCollection(
                {
                    "val/dice": Dice(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None),
                    "val/F1": F1(num_classes=self.num_classes, ignore_index=0 if self.num_classes > 1 else None, average=None),
                },
            )
            _default_loss = "DiceCE"

        if self.task_type == "self-supervised":
            self.train_metrics = MetricCollection({"train/MAE": MeanAbsoluteError()})
            self.val_metrics = MetricCollection({"train/MAE": MeanAbsoluteError()})
            _default_loss = "MSE"

        if self.loss_fn is None:
            self.loss_fn = _default_loss

        self.log_image_every_n_epochs = 1
        # Inference
        self.sliding_window_overlap = sliding_window_overlap
        self.test_time_augmentation = test_time_augmentation

        # If we are training we save params and then start training
        # Do not overwrite parameters during inference.
        self.save_hyperparameters()
        self.load_model()

    def load_model(self):
        logging.info(f"Loading Model: {self.model_dimensions} {self.model_name}")
        self.model = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "network_architectures")],
            class_name=self.model_name,
            current_module="yucca.network_architectures",
        )
        if self.model_dimensions == "3D":
            conv_op = torch.nn.Conv3d
            norm_op = torch.nn.InstanceNorm3d
        else:
            conv_op = torch.nn.Conv2d
            norm_op = torch.nn.BatchNorm2d

        model_kwargs = {
            # Applies to all models
            "input_channels": self.num_modalities,
            "num_classes": self.num_classes,
            # Applies to most CNN-based architectures
            "conv_op": conv_op,
            "deep_supervision": self.deep_supervision,
            "norm_op": norm_op,
            # UNetR
            "patch_size": self.patch_size,
            # MedNeXt
            "checkpoint_style": "outside_block",
        }
        model_kwargs = filter_kwargs(self.model, model_kwargs)

        self.model = self.model(**model_kwargs)

    def forward(self, inputs):
        return self.model(inputs)

    def teardown(self, stage: str):  # noqa: U100
        wandb.finish()

    def training_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_train(output, target)

        if self.deep_supervision:
            # If deep_supervision is enabled output and target will be a list of (downsampled) tensors.
            # We only need the original ground truth and its corresponding prediction which is always the first entry in each list.
            output = output[0]
            target = target[0]

        metrics = self.compute_metrics(self.train_metrics, output, target)
        self.log_dict(
            {"train/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and self.current_epoch % self.log_image_every_n_epochs == 0 and wandb.run is not None:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="train",
                task_type=self.task_type,
            )

        return loss

    def validation_step(self, batch, batch_idx):
        inputs, target, file_path = batch["image"], batch["label"], batch["file_path"]
        output = self(inputs)
        loss = self.loss_fn_val(output, target)
        metrics = self.compute_metrics(self.val_metrics, output, target)
        self.log_dict(
            {"val/loss": loss} | metrics,
            on_step=self.step_logging,
            on_epoch=self.epoch_logging,
            prog_bar=self.progress_bar,
            logger=True,
        )

        if batch_idx == 0 and self.current_epoch % self.log_image_every_n_epochs == 0 and wandb.run is not None:
            self._log_dict_of_images_to_wandb(
                {
                    "input": inputs.detach().cpu().to(torch.float32).numpy(),
                    "target": target.detach().cpu().to(torch.float32).numpy(),
                    "output": output.detach().cpu().to(torch.float32).numpy(),
                    "file_path": file_path,
                },
                log_key="val",
                task_type=self.task_type,
            )

    def on_predict_start(self):
        preprocessor_class = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "preprocessing")],
            class_name=self.plans["preprocessor"],
            current_module="yucca.preprocessing",
        )
        self.preprocessor = preprocessor_class(join(self.version_dir, "hparams.yaml"))

    def predict_step(self, batch, _batch_idx, _dataloader_idx=0):
        case, case_id = batch
        (
            case_preprocessed,
            case_properties,
        ) = self.preprocessor.preprocess_case_for_inference(case, self.patch_size, self.sliding_window_prediction)
        logits = self.model.predict(
            data=case_preprocessed,
            mode=self.model_dimensions,
            mirror=self.test_time_augmentation,
            overlap=self.sliding_window_overlap,
            patch_size=self.patch_size,
            sliding_window_prediction=self.sliding_window_prediction,
        )

        logits, case_properties = self.preprocessor.reverse_preprocessing(logits, case_properties)
        return {"logits": logits, "properties": case_properties, "case_id": case_id[0]}

    def compute_metrics(self, metrics, output, target, ignore_index: int = 0):
        metrics = metrics(output, target)
        tmp = {}
        to_drop = []
        for key in metrics.keys():
            if metrics[key].numel() > 1:
                to_drop.append(key)
                for i, val in enumerate(metrics[key]):
                    if not i == ignore_index:
                        tmp[key + "_" + str(i)] = val
        for k in to_drop:
            metrics.pop(k)
        metrics.update(tmp)
        return metrics

    def configure_optimizers(self):
        # Initialize and configure the loss(es) here.
        # loss_kwargs holds args for any scheduler class,
        # but using filtering we only pass arguments relevant to the selected class.
        self.loss_fn = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "training/loss_and_optim")],
            class_name=self.loss_fn,
            current_module="yucca.training.loss_and_optim",
        )
        loss_kwargs = {
            # DCE
            "soft_dice_kwargs": {"apply_softmax": True},
        }

        loss_kwargs = filter_kwargs(self.loss_fn, loss_kwargs)

        self.loss_fn_train = self.loss_fn(**loss_kwargs)
        self.loss_fn_val = self.loss_fn(**loss_kwargs)

        # If deep_supervision is enabled we wrap our training loss (and potentially specify weights)
        # We leave the validation loss as is, as deep_supervision is not used for validation.
        if self.deep_supervision:
            self.loss_fn_train = DeepSupervisionLoss(self.loss_fn_train, weights=None)

        # Initialize and configure the optimizer(s) here.
        # optim_kwargs holds args for any scheduler class,
        # but using filtering we only pass arguments relevant to the selected class.
        optim_kwargs = {
            # all
            "lr": self.lr,
            # SGD
            "momentum": self.momentum,
            "eps": 1e-4,
            "weight_decay": 3e-5,
        }

        optim_kwargs = filter_kwargs(self.optim, optim_kwargs)

        self.optim = self.optim(self.model.parameters(), **optim_kwargs)

        # Initialize and configure LR scheduler(s) here
        # lr_scheduler_kwargs holds args for any scheduler class,
        # but using filtering we only pass arguments relevant to the selected class.
        lr_scheduler_kwargs = {
            # Cosine Annealing
            "T_max": self.trainer.max_epochs,
            "eta_min": 1e-9,
        }

        lr_scheduler_kwargs = filter_kwargs(self.lr_scheduler, lr_scheduler_kwargs)

        self.lr_scheduler = self.lr_scheduler(self.optim, **lr_scheduler_kwargs)

        # Finally return the optimizer and scheduler - the loss is not returned.
        return {"optimizer": self.optim, "lr_scheduler": self.lr_scheduler}

    def load_state_dict(self, state_dict, *args, **kwargs):
        # First we filter out layers that have changed in size
        # This is often the case in the output layer.
        # If we are finetuning on a task with a different number of classes
        # than the pretraining task, the # output channels will have changed.
        old_params = copy.deepcopy(self.state_dict())
        state_dict = {
            k: v for k, v in state_dict.items() if (k in old_params) and (old_params[k].shape == state_dict[k].shape)
        }
        rejected_keys_new = [k for k in state_dict.keys() if k not in old_params]
        rejected_keys_shape = [k for k in state_dict.keys() if old_params[k].shape != state_dict[k].shape]
        rejected_keys_data = []

        # Here there's also potential to implement custom loading functions.
        # E.g. to load 2D pretrained models into 3D by repeating or something like that.

        # Now keep track of the # of layers with succesful weight transfers
        successful = 0
        unsuccessful = 0
        super().load_state_dict(state_dict, *args, **kwargs)
        new_params = self.state_dict()
        for param_name, p1, p2 in zip(old_params.keys(), old_params.values(), new_params.values()):
            # If more than one param in layer is NE (not equal) to the original weights we've successfully loaded new weights.
            if p1.data.ne(p2.data).sum() > 0:
                successful += 1
            else:
                unsuccessful += 1
                if param_name not in rejected_keys_new and param_name not in rejected_keys_shape:
                    rejected_keys_data.append(param_name)

        logging.warn(f"Succesfully transferred weights for {successful}/{successful+unsuccessful} layers")
        logging.warn(
            f"Rejected the following keys:\n"
            f"Not in old dict: {rejected_keys_new}.\n"
            f"Wrong shape: {rejected_keys_shape}.\n"
            f"Post check not succesful: {rejected_keys_data}."
        )

        return successful

    def _log_dict_of_images_to_wandb(self, imagedict: {}, log_key: str, task_type: str = "segmentation"):
        # This needs to handle the following cases:
        # Segmentation      : {"input": (b,m,x,y(,z)), "target": (b,1,x,y(,z)), "output": (b,c,x,y(,z)), "file_path": [pathA, pathB, ...]}
        # Self-supervised   : {"input": (b,m,x,y(,z)), "target": (b,m,x,y(,z)), "output": (b,m,x,y(,z)), "file_path": [pathA, pathB, ...]}
        # Classification    : {"input": (b,m,x,y(,z)), "target": (b,1,x), "output": (b,c,x), "file_path": [pathA, pathB, ...]}

        batch_idx = np.random.randint(0, imagedict["input"].shape[0])
        channel_idx = np.random.randint(0, imagedict["input"].shape[1])

        if len(imagedict["input"].shape) == 5:  # 3D images.
            # We need to select a slice to visualize.
            if task_type == "segmentation" and len(imagedict["target"][batch_idx, 0].nonzero()[0]) > 0:
                # Select a foreground slice if any exist.
                foreground_locations = imagedict["target"][batch_idx, 0].nonzero()
                slice_to_visualize = foreground_locations[0][np.random.randint(0, len(foreground_locations[0]))]
            else:
                slice_to_visualize = np.random.randint(0, imagedict["input"].shape[2])

            imagedict["input"] = imagedict["input"][:, :, slice_to_visualize]
            if len(imagedict["target"].shape) == 5:
                imagedict["target"] = imagedict["target"][:, :, slice_to_visualize]
            if len(imagedict["output"].shape) == 5:
                imagedict["output"] = imagedict["output"][:, :, slice_to_visualize]

        image = imagedict["input"][batch_idx, channel_idx]
        case = os.path.splitext(os.path.split(imagedict["file_path"][batch_idx])[-1])[0]

        if task_type in ["segmentation", "classification"]:
            target = imagedict["target"][batch_idx, 0]
            output = imagedict["output"][batch_idx].argmax(0)
        elif task_type == "self-supervised":
            target = imagedict["target"][batch_idx, channel_idx]
            output = imagedict["output"][batch_idx, channel_idx]
        else:
            logging.warn(
                f"Unknown task type. Found {task_type} and expected one in ['classification', 'segmentation', 'self-supervised']"
            )

        if len(target.shape) == 1:
            fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(5, 5), dpi=100, constrained_layout=True)
            axes[0].imshow(image, cmap="gray", vmin=np.quantile(image, 0.01), vmax=np.quantile(image, 0.99))
            axes[0].set_title("input")
            fig.suptitle(f"{case}. Target: {target} | Output: {output}", fontsize=16)
        else:
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(15, 5), dpi=100, constrained_layout=True)
            axes[0].imshow(image, cmap="gray", vmin=np.quantile(image, 0.01), vmax=np.quantile(image, 0.99))
            axes[0].set_title("input")
            axes[1].imshow(target, cmap="gray")
            axes[1].set_title("target")
            axes[2].imshow(output, cmap="gray")
            axes[2].set_title("output")
            fig.suptitle(case, fontsize=16)

            wandb.log({log_key: wandb.Image(fig)}, commit=False)
            plt.close(fig)



import numpy as np
from batchgenerators.utilities.file_and_folder_operations import load_pickle


class YuccaLoader(object):
    """
    Default data loader for patch-based dataloading and augmentation.

    This class takes input data in 2D or 3D, batch size, generator patch size and
    the probability to oversample foreground. If the data is 3D it outputs either 2D or 3D data,
    depending on the model. If the data is 2D it outputs 2D data.

    3D data will be of the shape (b, c, x, y, z) and 2D data will be shape (b, c, x, y)
    with b = batch size, c = modalities (derived from the modalities of the input data) and
    xy(z) defined by the generator patch size.

    The generator patch size is larger than the final patch size to remove border interpolation
    artifacts. If we cropped to the final patch size here, and then performed transformations such
    as rotations or scaling we may get large border artifacts or black areas where there could
    have been parts of the real image, if we started out with a larger area then did the
    augmentations and THEN cropped to the desired size.

    This class also supports set_thread_id and __next__ functions required for the CPU threading and
    background implementation used to apply augmentations in the framework.

    Finally, by default it will keep cases in RAM to save the time required to repeatedly load the
    image, however this should be avoided for very large datasets.
    """

    def __init__(
        self,
        list_of_files,
        batch_size,
        gen_patch_size,
        p_oversample_foreground=0.5,
        keep_in_ram=True,
    ):
        self.batch_size = batch_size
        self.gen_patch_size = gen_patch_size
        self.files = list_of_files
        self.p_oversample_foreground = p_oversample_foreground
        self.keep_in_ram = keep_in_ram
        self.already_loaded_cases = {}
        self.determine_shapes()

    def determine_shapes(self):
        image = self.load_and_maybe_keep_volume(self.files[0])
        self.input_shape = image.shape
        self.image_shape_for_aug = (
            self.batch_size,
            image.shape[0] - 1,
            *self.gen_patch_size,
        )
        self.seg_shape_for_aug = (self.batch_size, 1, *self.gen_patch_size)
        del image

    def generate_train_batch(self):
        if len(self.gen_patch_size) == 3:
            return self.generate_3D_batch()
        if len(self.gen_patch_size) == 2 and len(self.input_shape) == 4:
            return self.generate_2D_batch_from_3D()
        if len(self.gen_patch_size) == 2 and len(self.input_shape) == 3:
            return self.generate_2D_batch_from_2D()
        else:
            print(f"patch size should be (x, y, z) or (x,y) but is: {self.gen_patch_size}")

    def generate_3D_batch(self):
        """
        The possible input for this can be 2D or 3D data.
        For 2D we want to pad or crop as necessary.
        For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
        then pad or crop as necessary.
        """
        image_batch = np.zeros(self.image_shape_for_aug)
        seg_batch = np.zeros(self.seg_shape_for_aug)

        samples = np.random.choice(self.files, self.batch_size)

        for idx, sample in enumerate(samples):
            imseg = self.load_and_maybe_keep_volume(sample)
            image_properties = self.load_and_maybe_keep_pickle(sample[:-4] + ".pkl")

            assert len(imseg.shape) == 4, "input should be (c, x, y, z)" f" but it is: {imseg.shape}"

            # First we pad to ensure min size is met
            to_pad = []
            for d in range(3):
                if imseg.shape[d + 1] < self.gen_patch_size[d]:
                    to_pad += [(self.gen_patch_size[d] - imseg.shape[d + 1])]
                else:
                    to_pad += [0]

            pad_lb_x = to_pad[0] // 2
            pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
            pad_lb_y = to_pad[1] // 2
            pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2
            pad_lb_z = to_pad[2] // 2
            pad_ub_z = to_pad[2] // 2 + to_pad[2] % 2

            # This is where we should implement any patch selection biases.
            # The final patch excted after augmentation will always be the center of this patch
            # as this is where artefacts are least present
            crop_start_idx = []
            if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
                for d in range(3):
                    if imseg.shape[d + 1] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [np.random.randint(imseg.shape[d + 1] - self.gen_patch_size[d] + 1)]
            else:
                locidx = np.random.choice(len(image_properties["foreground_locations"]))
                location = image_properties["foreground_locations"][locidx]
                for d in range(3):
                    if imseg.shape[d + 1] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [
                            np.random.randint(
                                max(0, location[d] - self.gen_patch_size[d]),
                                min(
                                    location[d],
                                    imseg.shape[d + 1] - self.gen_patch_size[d],
                                )
                                + 1,
                            )
                        ]

            image_batch[
                idx,
                :,
                :,
                :,
                :,
            ] = np.pad(
                imseg[
                    :-1,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                    crop_start_idx[2] : crop_start_idx[2] + self.gen_patch_size[2],
                ],
                (
                    (0, 0),
                    (pad_lb_x, pad_ub_x),
                    (pad_lb_y, pad_ub_y),
                    (pad_lb_z, pad_ub_z),
                ),
                mode="edge",
            )

            seg_batch[
                idx,
                :,
                :,
                :,
                :,
            ] = np.pad(
                imseg[
                    -1:,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                    crop_start_idx[2] : crop_start_idx[2] + self.gen_patch_size[2],
                ],
                (
                    (0, 0),
                    (pad_lb_x, pad_ub_x),
                    (pad_lb_y, pad_ub_y),
                    (pad_lb_z, pad_ub_z),
                ),
            )

        data_dict = {"image": image_batch, "seg": seg_batch}
        return data_dict

    def generate_2D_batch_from_3D(self):
        """
        The possible input for this can be 2D or 3D data.
        For 2D we want to pad or crop as necessary.
        For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
        then pad or crop as necessary.
        """
        image_batch = np.zeros(self.image_shape_for_aug)
        seg_batch = np.zeros(self.seg_shape_for_aug)

        samples = np.random.choice(self.files, self.batch_size)

        for idx, sample in enumerate(samples):
            imseg = self.load_and_maybe_keep_volume(sample)
            image_properties = self.load_and_maybe_keep_pickle(sample[:-4] + ".pkl")

            assert len(imseg.shape) == 4, "input should be (c, x, y, z)" f" but it is: {imseg.shape}"

            # First we pad to ensure min size is met
            to_pad = []
            for d in range(2):
                if imseg.shape[d + 2] < self.gen_patch_size[d]:
                    to_pad += [(self.gen_patch_size[d] - imseg.shape[d + 2])]
                else:
                    to_pad += [0]

            pad_lb_y = to_pad[0] // 2
            pad_ub_y = to_pad[0] // 2 + to_pad[0] % 2
            pad_lb_z = to_pad[1] // 2
            pad_ub_z = to_pad[1] // 2 + to_pad[1] % 2

            # This is where we should implement any patch selection biases.
            # The final patch extracted after augmentation will always be the center of this patch
            # as this is where augmentation-induced interpolation artefacts are least likely
            crop_start_idx = []
            if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
                x_idx = np.random.randint(imseg.shape[1])
                for d in range(2):
                    if imseg.shape[d + 2] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [np.random.randint(imseg.shape[d + 2] - self.gen_patch_size[d] + 1)]
            else:
                locidx = np.random.choice(len(image_properties["foreground_locations"]))
                location = image_properties["foreground_locations"][locidx]
                x_idx = location[0]
                for d in range(2):
                    if imseg.shape[d + 2] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [
                            np.random.randint(
                                max(0, location[d + 1] - self.gen_patch_size[d]),
                                min(
                                    location[d + 1],
                                    imseg.shape[d + 2] - self.gen_patch_size[d],
                                )
                                + 1,
                            )
                        ]

            image_batch[idx, :, :, :] = np.pad(
                imseg[
                    :-1,
                    x_idx,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                ],
                ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
                mode="edge",
            )
            seg_batch[idx, :, :, :] = np.pad(
                imseg[
                    -1:,
                    x_idx,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                ],
                ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
            )

        data_dict = {"image": image_batch, "seg": seg_batch}
        return data_dict

    def generate_2D_batch_from_2D(self):
        """
        The possible input for this can be 2D or 3D data.
        For 2D we want to pad or crop as necessary.
        For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
        then pad or crop as necessary.
        """
        image_batch = np.zeros(self.image_shape_for_aug)
        seg_batch = np.zeros(self.seg_shape_for_aug)

        samples = np.random.choice(self.files, self.batch_size)

        for idx, sample in enumerate(samples):
            imseg = self.load_and_maybe_keep_volume(sample)
            image_properties = self.load_and_maybe_keep_pickle(sample[:-4] + ".pkl")

            assert len(imseg.shape) == 3, "input should be (c, x, y)" f" but it is: {imseg.shape}"

            # First we pad to ensure min size is met
            to_pad = []
            for d in range(2):
                if imseg.shape[d + 1] < self.gen_patch_size[d]:
                    to_pad += [(self.gen_patch_size[d] - imseg.shape[d + 1])]
                else:
                    to_pad += [0]

            pad_lb_x = to_pad[0] // 2
            pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
            pad_lb_y = to_pad[1] // 2
            pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2

            # This is where we should implement any patch selection biases.
            # The final patch extracted after augmentation will always be the center of this patch
            # as this is where artefacts are least present
            crop_start_idx = []
            if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= self.p_oversample_foreground:
                for d in range(2):
                    if imseg.shape[d + 1] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [np.random.randint(imseg.shape[d + 1] - self.gen_patch_size[d] + 1)]
            else:
                locidx = np.random.choice(len(image_properties["foreground_locations"]))
                location = image_properties["foreground_locations"][locidx]
                for d in range(2):
                    if imseg.shape[d + 1] < self.gen_patch_size[d]:
                        crop_start_idx += [0]
                    else:
                        crop_start_idx += [
                            np.random.randint(
                                max(0, location[d] - self.gen_patch_size[d]),
                                min(
                                    location[d],
                                    imseg.shape[d + 1] - self.gen_patch_size[d],
                                )
                                + 1,
                            )
                        ]

            image_batch[idx, :, :, :] = np.pad(
                imseg[
                    :-1,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                ],
                ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
                mode="edge",
            )

            seg_batch[idx, :, :, :] = np.pad(
                imseg[
                    -1:,
                    crop_start_idx[0] : crop_start_idx[0] + self.gen_patch_size[0],
                    crop_start_idx[1] : crop_start_idx[1] + self.gen_patch_size[1],
                ],
                ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
            )
        data_dict = {"image": image_batch, "seg": seg_batch}
        return data_dict

    def __next__(self):
        return self.generate_train_batch()

    def set_thread_id(self, thread_id):
        self.thread_id = thread_id

    def load_and_maybe_keep_pickle(self, picklepath):
        if not self.keep_in_ram:
            return load_pickle(picklepath)
        if picklepath in self.already_loaded_cases:
            return self.already_loaded_cases[picklepath]
        self.already_loaded_cases[picklepath] = load_pickle(picklepath)
        return self.already_loaded_cases[picklepath]

    def load_and_maybe_keep_volume(self, path):
        if not self.keep_in_ram:
            if path[-3:] == "npy":
                return np.load(path, "r")
            image = np.load(path)
            assert len(image.files) == 1, (
                "More than one entry in data array. " f"Should only be ['data'] but is {[key for key in image.files]}"
            )
            return image[image.files[0]]

        if path in self.already_loaded_cases:
            return self.already_loaded_cases[path]

        if path[-3:] == "npy":
            try:
                self.already_loaded_cases[path] = np.load(path, "r")
            except ValueError:
                self.already_loaded_cases[path] = np.load(path, allow_pickle=True)
            return self.already_loaded_cases[path]

        image = np.load(path)
        assert len(image.files) == 1, (
            "More than one entry in data array. " f"Should only be ['data'] but is {[key for key in image.files]}"
        )
        self.already_loaded_cases = image[image.files[0]]
        return self.already_loaded_cases[path]



from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data
import shutil
from tqdm import tqdm
from os.path import basename


def convert(_path: str):
    # This script assumes TC has already been done on supervised version of dataset with
    # name `Task012_BratS21`

    # Target names
    sup_task_name = "Task012_BraTS21"
    ssl_task_name = "Task209_BraTS21"

    source_folder = join(yucca_raw_data, sup_task_name, "imagesTr")
    target_base = join(yucca_raw_data, ssl_task_name)

    target_imagesTr = join(target_base, "imagesTr")
    maybe_mkdir_p(target_imagesTr)

    suffix = ".nii.gz"
    training_samples = subfiles(source_folder, suffix=suffix)

    ###Populate Target Directory###
    for src_path in tqdm(training_samples):
        file_name = basename(src_path)
        img_name = file_name[:-11]  # removing last _00x_.nii.gz

        if "_000." in file_name:
            dest_path = f"{target_imagesTr}/{img_name}_flair_000.nii.gz"
        elif "_001." in file_name:
            dest_path = f"{target_imagesTr}/{img_name}_t1_000.nii.gz"
        elif "_002." in file_name:
            dest_path = f"{target_imagesTr}/{img_name}_t1ce_000.nii.gz"
        elif "_003." in file_name:
            dest_path = f"{target_imagesTr}/{img_name}_t2_000.nii.gz"
        else:
            raise ValueError

        shutil.copy2(src_path, dest_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=("MRI"),
        labels=None,
        dataset_name=sup_task_name,
        license="hands off!",
        dataset_description="BraTS21",
        dataset_reference="https://www.nitrc.org/projects/msseg, https://arxiv.org/abs/2206.06694",
    )



import argparse
import yucca
from yucca.utils.task_ids import maybe_get_task_from_task_id
from yucca.utils.files_and_folders import recursive_find_python_class
from batchgenerators.utilities.file_and_folder_operations import join


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-t", "--task", help="Name of the task to preprocess. " "Should be of format: TaskXXX_MYTASK", required=True
    )
    parser.add_argument(
        "-pl", help="Experiment Planner Class to employ. " "Defaults to the YuccaPlanner", default="YuccaPlanner"
    )
    parser.add_argument(
        "-pr",
        help="Preprocessor Class to employ. "
        "Defaults to the YuccaPreprocessor, but can be ClassificationPreprocessor for classification tasks",
        default="YuccaPreprocessor",
    )
    parser.add_argument(
        "-v",
        help="Designate target view or orientation to obtain with transposition. "
        "Standard settings will handle this for you, but use this to manually specify. "
        "Can be 'X', 'Y' or 'Z'",
    )
    parser.add_argument(
        "--ensemble",
        help="Used to initialize data preprocessing for ensemble/2.5D training",
        default=False,
        action="store_true",
    )
    parser.add_argument("--disable_sanity_checks", help="Enable or disable sanity checks", default=False)
    parser.add_argument("--threads", help="Used to specify the number of processes to use for preprocessing", default=2)
    args = parser.parse_args()

    task = maybe_get_task_from_task_id(args.task)
    planner_name = args.pl
    preprocessor_name = args.pr
    view = args.v
    disable_sanity_checks = args.disable_sanity_checks
    ensemble = args.ensemble
    threads = args.threads

    if not ensemble:
        planner = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "planning")], class_name=planner_name, current_module="yucca.planning"
        )
        planner = planner(task, preprocessor_name, threads=threads, disable_sanity_checks=disable_sanity_checks, view=view)
        planner.plan()
        planner.preprocess()
    if ensemble:
        views = ["X", "Y", "Z"]
        for view in views:
            planner = recursive_find_python_class(
                folder=[join(yucca.__path__[0], "planning")], class_name=planner_name, current_module="yucca.planning"
            )
            planner = planner(task, preprocessor_name, threads=threads, disable_sanity_checks=disable_sanity_checks, view=view)
            planner.plan()
            planner.preprocess()


if __name__ == "__main__":
    main()



import numpy as np
from scipy.ndimage import gaussian_filter
import math as m


def create_zero_centered_coordinate_matrix(shape):
    if len(shape) == 3:
        mesh = np.array(
            np.meshgrid(
                np.arange(shape[0]),
                np.arange(shape[1]),
                np.arange(shape[2]),
                indexing="ij",
            )
        ).astype(float)
    if len(shape) == 2:
        mesh = np.array(np.meshgrid(np.arange(shape[0]), np.arange(shape[1]), indexing="ij")).astype(float)

    for d in range(len(shape)):
        mesh[d] -= (mesh.shape[d + 1] - 1) / 2
        assert np.mean(mesh[d]) == 0, "beware: mesh didnt zero-center"
    return mesh


def deform_coordinate_matrix(coordinate_matrix, alpha, sigma):
    deforms = np.array(
        [
            gaussian_filter(
                (np.random.random(coordinate_matrix.shape[1:]) * 2 - 1),
                sigma,
                mode="constant",
                cval=0,
            )
            * alpha
            for _ in range(coordinate_matrix.shape[0])
        ]
    )
    coordinate_matrix = deforms + coordinate_matrix
    return coordinate_matrix


def Rx(theta):
    return np.array([[1, 0, 0], [0, m.cos(theta), -m.sin(theta)], [0, m.sin(theta), m.cos(theta)]])


def Ry(theta):
    return np.array([[m.cos(theta), 0, m.sin(theta)], [0, 1, 0], [-m.sin(theta), 0, m.cos(theta)]])


def Rz(theta):
    return np.array([[m.cos(theta), -m.sin(theta), 0], [m.sin(theta), m.cos(theta), 0], [0, 0, 1]])


def Rz2D(theta):
    return np.array([[m.cos(theta), -m.sin(theta)], [m.sin(theta), m.cos(theta)]])


def get_max_rotated_size(patch_size):
    if len(patch_size) == 2:
        max_dim = int(np.sqrt(patch_size[0] ** 2 + patch_size[1] ** 2))
        return (max_dim, max_dim)

    max_dim_0 = max(
        int(np.sqrt(patch_size[0] ** 2 + patch_size[1] ** 2)),
        int(np.sqrt(patch_size[0] ** 2 + patch_size[2] ** 2)),
    )

    max_dim_1 = max(
        int(np.sqrt(patch_size[1] ** 2 + patch_size[0] ** 2)),
        int(np.sqrt(patch_size[1] ** 2 + patch_size[2] ** 2)),
    )

    max_dim_2 = max(
        int(np.sqrt(patch_size[2] ** 2 + patch_size[0] ** 2)),
        int(np.sqrt(patch_size[2] ** 2 + patch_size[1] ** 2)),
    )

    return (max_dim_0, max_dim_1, max_dim_2)



import numpy as np
from yucca.utils.nib_utils import get_nib_spacing
from yucca.evaluation.deepmind_surface_distance import metrics


def get_surface_metrics_for_label(gt, pred, label, as_binary: bool = False):
    spacing = get_nib_spacing(pred)
    pred = pred.get_fdata()
    gt = gt.get_fdata()
    labeldict = {}

    if label == 0:
        labeldict["Average Surface Distance"] = 0
        return labeldict
    if as_binary:
        gt = gt.astype(bool)
        pred = pred.astype(bool)
    else:
        pred = np.where(pred == label, 1, 0).astype(bool)
        gt = np.where(gt == label, 1, 0).astype(bool)

    surface_distances = metrics.compute_surface_distances(
        mask_gt=gt,
        mask_pred=pred,
        spacing_mm=spacing,
    )

    labeldict["Average Surface Distance"] = metrics.compute_surface_dice_at_tolerance(
        surface_distances=surface_distances, tolerance_mm=1
    )
    return labeldict



import os
from yucca.task_conversion.utils import combine_imagesTr_from_tasks, generate_dataset_json
from yucca.paths import yucca_raw_data


# Define the name of the new task
task_name = "TaskXXX_MyCombinedTask"

# Define the expected labels (leave empty for no labels)
expected_labels = {}

# Define expected modalities, such as ('MRI', 'CT',) or just ('Medical Images',) - remember trailing "," for singular modalities
expected_modalities = ("Unspecified",)

# Define the tasks to combine, such as ["Task001_OASIS", "Task002_LPBA40"]
# The individual task_conversion scripts must be run prior to executing this, as the script will look for the data in the yucca_raw_data folder.
tasks_to_combine = []

### In most cases the remaining can be left untouched ###
# Setting the paths to save the new task and making the directories
target_base = os.path.join(yucca_raw_data, task_name)
target_imagesTr = os.path.join(yucca_raw_data, task_name, "imagesTr")
target_imagesTs = None
os.makedirs(target_imagesTr, exist_ok=True)

combine_imagesTr_from_tasks(tasks=tasks_to_combine, task_name=task_name, target_dir=target_imagesTr)

generate_dataset_json(
    os.path.join(target_base, "dataset.json"),
    target_imagesTr,
    target_imagesTs,
    modalities=expected_modalities,
    labels=expected_labels,
    dataset_name=task_name,
    license="CC-BY-SA 4.0",
    dataset_description="",
    dataset_reference="",
)



import logging
from lightning.pytorch.loggers.logger import Logger
from typing import Optional, Union
from dataclasses import dataclass
from lightning.pytorch.callbacks import ModelCheckpoint, LearningRateMonitor
from lightning.pytorch.profilers import AdvancedProfiler
from lightning.pytorch.loggers import WandbLogger
from yucca.evaluation.loggers import YuccaLogger
from yucca.utils.saving import WritePredictionFromLogits
from yucca.paths import yucca_wandb_entity
from lightning.pytorch.profilers.profiler import Profiler


@dataclass
class CallbackConfig:
    callbacks: list
    loggers: list
    profiler: Profiler
    wandb_id: str

    def lm_hparams(self):
        # A story of a shitty bug 🐞
        # When using DDP `self.wandb_id` will on any RANK ≠ 0 device be set to something in the style of
        # <bound method _DummyExperiment.nop of <lightning.fabric.loggers.logger._DummyExperiment object at 0x7fb5a17c65d0>>
        # which will make self.save_hyperparameters() crash on that device, because (ofc) the above is not serializable.
        # So we cast it to a str, which makes it serializable, and since checkpoints are only stored from the
        # rank zero device, we will never see this giberish again.
        return {"wandb_id": str(self.wandb_id)}


def get_callback_config(
    save_dir: str,
    version_dir: str,
    version: int,
    ckpt_version_dir: Optional[str] = None,
    ckpt_wandb_id: Optional[str] = None,
    enable_logging: bool = True,
    interval_ckpt_epochs: int = 250,
    latest_ckpt_epochs: int = 25,
    log_lr: bool = True,
    log_model: Union[str, bool] = True,
    prediction_output_dir: Optional[str] = None,
    profile: bool = False,
    project: str = "Yucca",
    save_softmax: bool = False,
    steps_per_epoch: int = 250,
    store_best_ckpt: bool = True,
    write_predictions: bool = True,
    run_name: Optional[str] = None,
    run_description: Optional[str] = None,
    experiment: Optional[str] = None,
    # An entity is a username or team name where you're sending runs.
    # If not provided will send to default entity, usually username.
    wandb_entity: Optional[str] = None,
):
    callbacks = get_callbacks(
        interval_ckpt_epochs,
        latest_ckpt_epochs,
        prediction_output_dir,
        save_softmax,
        write_predictions,
        store_best_ckpt,
        log_lr,
    )
    loggers = get_loggers(
        ckpt_version_dir=ckpt_version_dir,
        ckpt_wandb_id=ckpt_wandb_id,
        save_dir=save_dir,
        version_dir=version_dir,
        version=version,
        enable_logging=enable_logging,
        steps_per_epoch=steps_per_epoch,
        project=project,
        log_model=log_model,
        run_name=run_name,
        run_description=run_description,
        experiment=experiment,
        wandb_entity=wandb_entity,
    )
    wandb_id = get_wandb_id(loggers, enable_logging)
    profiler = get_profiler(profile, version_dir)

    return CallbackConfig(callbacks=callbacks, loggers=loggers, profiler=profiler, wandb_id=wandb_id)


def get_loggers(
    ckpt_version_dir: Union[str, None],
    ckpt_wandb_id: Union[str, None],
    enable_logging: bool,
    log_model: Union[bool, str],
    project: str,
    save_dir: str,
    steps_per_epoch: int,
    version_dir: str,
    version: Union[int, str],
    run_name: str,
    run_description: str,
    experiment: str,
    wandb_entity: Optional[str],
):
    # The YuccaLogger is the barebones logger needed to save hparams.yaml
    # It should generally never be disabled.
    if isinstance(version, str):
        version = int(version)

    if wandb_entity is None:
        wandb_entity = yucca_wandb_entity

    loggers = [
        YuccaLogger(
            disable_logging=not enable_logging,
            save_dir=save_dir,
            name=None,
            version=version,
            steps_per_epoch=steps_per_epoch,
        )
    ]
    if enable_logging:
        use_ckpt_id = should_use_ckpt_wandb_id(ckpt_version_dir, ckpt_wandb_id, version_dir)
        loggers.append(
            WandbLogger(
                name=run_name or f"{experiment}_V{version}",
                notes=run_description,
                save_dir=version_dir,
                project=project,
                group=experiment,
                log_model=log_model,
                version=ckpt_wandb_id if use_ckpt_id else None,
                resume="allow" if use_ckpt_id else None,
                entity=wandb_entity,
            )
        )

    return loggers


def get_callbacks(
    interval_ckpt_every_n_epochs: int,
    last_ckpt_every_n_epochs: int,
    prediction_output_dir: str,
    save_softmax: bool,
    write_predictions: bool,
    store_best_ckpt: bool,
    log_lr: bool,
):
    interval_ckpt = ModelCheckpoint(
        every_n_epochs=interval_ckpt_every_n_epochs, save_top_k=-1, filename="{epoch}", enable_version_counter=False
    )
    latest_ckpt = ModelCheckpoint(
        every_n_epochs=last_ckpt_every_n_epochs,
        save_top_k=1,
        filename="last",
        enable_version_counter=False,
    )

    callbacks = [interval_ckpt, latest_ckpt]

    if store_best_ckpt:
        best_ckpt = ModelCheckpoint(
            monitor="val/loss", mode="min", save_top_k=1, filename="best", enable_version_counter=False
        )
        callbacks.append(best_ckpt)

    if write_predictions:
        pred_writer = WritePredictionFromLogits(
            output_dir=prediction_output_dir, save_softmax=save_softmax, write_interval="batch"
        )
        callbacks.append(pred_writer)

    if log_lr:
        lr_monitor = LearningRateMonitor(logging_interval="epoch", log_momentum=True)
        callbacks.append(lr_monitor)

    return callbacks


def get_profiler(profile: bool, outpath: str):
    if profile:
        logging.info(f"Profiling run. Saving results in {outpath}")
        return AdvancedProfiler(dirpath=outpath, filename="advanced_profile")
    return None


def get_wandb_id(loggers: list[Logger], enable_logging: bool):
    if enable_logging:
        wandb_logger = loggers[-1]
        assert isinstance(wandb_logger, WandbLogger)
        return wandb_logger.experiment.id
    return None


def should_use_ckpt_wandb_id(ckpt_version_dir, ckpt_wandb_id, version_dir):
    # If no wandb_id was found we can not (and should not try) reuse it.
    if ckpt_wandb_id is None:
        return False
    # If it exists and our current output directory INCLUDING THE CURRENT VERSION is equal
    # to the previous output directory we can safely assume we're continuing an
    # interrupted run.
    return ckpt_version_dir == version_dir



from yucca.image_processing.transforms.YuccaTransform import YuccaTransform
import numpy as np
from skimage.transform import resize


class DownsampleSegForDS(YuccaTransform):
    """ """

    def __init__(self, deep_supervision: bool = False, label_key="label", factors=(1, 0.5, 0.25, 0.125, 0.0625)):
        self.deep_supervision = deep_supervision
        self.label_key = label_key
        self.factors = factors

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __downsample__(self, label, factors):
        orig_type = label.dtype
        orig_shape = label.shape
        downsampled_labels = []
        for factor in factors:
            target_shape = np.array(orig_shape).astype(int)
            for i in range(2, len(orig_shape)):
                target_shape[i] *= factor
            if np.all(target_shape == orig_shape):
                downsampled_labels.append(label)
            else:
                canvas = np.zeros(target_shape)
                for b in range(label.shape[0]):
                    for c in range(label[b].shape[0]):
                        canvas[b, c] = resize(
                            label[b, c].astype(float),
                            target_shape[2:],
                            0,
                            mode="edge",
                            clip=True,
                            anti_aliasing=False,
                        ).astype(orig_type)
                downsampled_labels.append(canvas)
        return downsampled_labels

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        if self.deep_supervision:
            data_dict[self.label_key] = self.__downsample__(data_dict[self.label_key], self.factors)
        return data_dict
